import numpy as np


def points_to_waveform(times: np.ndarray, amplitudes: np.ndarray, grad_raster_time: float):
    """
    1D interpolate amplitude values `amplitudes` at time indices `times` as per the gradient raster time
    `grad_raster_time` to generate a gradient waveform.

    Parameters
    ----------
    times : numpy.ndarray
        Time indices.
    amplitudes : numpy.ndarray
        Amplitude values at time indices `times`.
    grad_raster_time : float
        Gradient raster time.

    Returns
    -------
    waveform : numpy.ndarray
        Gradient waveform.
    """
    grd = np.arange(start=round(min(times) / grad_raster_time),
                    stop=round(max(times) / grad_raster_time)) * grad_raster_time
    waveform = np.interp(x=grd + grad_raster_time / 2, xp=times, fp=amplitudes)

    return waveform



import numpy as np

from pypulseq.opts import Opts


def calc_ramp(k0: np.ndarray, k_end: np.ndarray, system: Opts = Opts(), max_points: int = 500, max_grad=np.zeros(0),
              max_slew=np.zeros(0)):
    """
    Join the points `k0` and `k_end` in three-dimensional  k-space in minimal time, observing the gradient and slew limits
    (`max_grad` and `max_slew` respectively), and the gradient strength G0 before `k0[:, 1]` and Gend after
    `k_end[:, 1]`. In the context of a fixed gradient dwell time this is a discrete problem with an a priori unknown
    number of discretization steps. Therefore this method tries out the optimization with 0 steps, then 1 step, and so
    on, until  all conditions can be fulfilled, thus yielding a short connection.

    Parameters
    ----------
    k0 : numpy.ndarray
        Two preceding points in k-space. Shape is `[3, 2]`. From these points, the starting gradient will be calculated.
    k_end : numpy.ndarray
        Two following points in k-space. Shape is `[3, 2]`. From these points, the target gradient will be calculated.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    max_points : int, optional
        Maximum number of k-space points to be used in connecting `k0` and `k_end`. Default is 500.
    max_grad : float/list, optional
        Maximum total gradient strength. Either a single value or one value for each coordinate, of shape `[3, 1]`.
        Default is 0.
     max_slew : float/list, optional
        Maximum total slew rate. Either a single value or one value for each coordinate, of shape `[3, 1]`.
        Default is 0.

    Returns
    -------
    k_out : numpy.ndarray
        Connected k-space trajectory.
    success : bool
        Boolean flag indicating if `k0` and `k_end` were successfully joined.
    """

    def __inside_limits(grad, slew):
        if mode == 0:
            grad2 = np.sum(np.square(grad), axis=1)
            slew2 = np.sum(np.square(slew), axis=1)
            ok = np.all(np.max(grad2) <= np.square(max_grad)) and np.all(np.max(slew2) <= np.square(max_slew))
        else:
            ok = (np.sum(np.max(np.abs(grad), axis=1) <= max_grad) == 3) and (
                    np.sum(np.max(np.abs(slew), axis=1) <= max_slew) == 3)

        return ok

    def __joinleft0(k0, k_end, G0, G_end, use_points):
        if use_points == 0:
            G = np.stack((G0, (k_end - k0) / grad_raster, G_end)).T
            S = (G[:, 1:] - G[:, :-1]) / grad_raster

            k_out_left = np.zeros((3, 0))
            success = __inside_limits(G, S)

            return success, k_out_left

        dk = (k_end - k0) / (use_points + 1)
        kopt = k0 + dk
        Gopt = (kopt - k0) / grad_raster
        Sopt = (Gopt - G0) / grad_raster

        okGopt = np.sum(np.square(Gopt)) <= np.square(max_grad)
        okSopt = np.sum(np.square(Sopt)) <= np.square(max_slew)

        if okGopt and okSopt:
            k_left = kopt
        else:
            a = np.multiply(max_grad, grad_raster)
            b = np.multiply(max_slew, grad_raster ** 2)

            dkprol = G0 * grad_raster
            dkconn = dk - dkprol

            ksl = k0 + dkprol + dkconn / np.linalg.norm(dkconn) * b
            Gsl = (ksl - k0) / grad_raster
            okGsl = np.sum(np.square(Gsl)) <= np.square(max_grad)

            kgl = k0 + np.multiply(dk / np.linalg.norm(dk), a)
            Ggl = (kgl - k0) / grad_raster
            Sgl = (Ggl - G0) / grad_raster
            okSgl = np.sum(np.square(Sgl)) <= np.square(max_slew)

            if okGsl:
                k_left = ksl
            elif okSgl:
                k_left = kgl
            else:
                c = np.linalg.norm(dkprol)
                c1 = np.divide(np.square(a) - np.square(b) + np.square(c), (2 * c))
                h = np.sqrt(np.square(a) - np.square(c1))
                kglsl = k0 + np.multiply(c1, np.divide(dkprol, np.linalg.norm(dkprol)))
                projondkprol = (kgl * dkprol.T) * (dkprol / np.linalg.norm(dkprol))
                hdirection = kgl - projondkprol
                kglsl = kglsl + h * hdirection / np.linalg.norm(hdirection)
                k_left = kglsl

        success, k = __joinright0(k_left, k_end, (k_left - k0) / grad_raster, G_end, use_points - 1)
        if len(k) != 0:
            if len(k.shape) == 1:
                k = k.reshape((len(k), 1))
            if len(k_left.shape) == 1:
                k_left = k_left.reshape((len(k_left), 1))
            k_out_left = np.hstack((k_left, k))
        else:
            k_out_left = k_left

        return success, k_out_left

    def __joinleft1(k0, k_end, G0, G_end, use_points):
        if use_points == 0:
            G = np.stack((G0, (k_end - k0) / grad_raster, G_end))
            S = (G[:, 1:] - G[:, :-1]) / grad_raster

            k_out_left = np.zeros((3, 0))
            success = __inside_limits(G, S)

            return success, k_out_left

        k_left = np.zeros(3)

        dk = (k_end - k0) / (use_points + 1)
        kopt = k0 + dk
        Gopt = (kopt - k0) / grad_raster
        Sopt = (Gopt - G0) / grad_raster

        okGopt = np.abs(Gopt) <= max_grad
        okSopt = np.abs(Sopt) <= max_slew

        dkprol = G0 * grad_raster
        dkconn = dk - dkprol

        ksl = k0 + dkprol + np.multiply(np.sign(dkconn), max_slew) * grad_raster ** 2
        Gsl = (ksl - k0) / grad_raster
        okGsl = np.abs(Gsl) <= max_grad

        kgl = k0 + np.multiply(np.sign(dk), max_grad) * grad_raster ** 2
        Ggl = (kgl - k0) / grad_raster
        Sgl = (Ggl - G0) / grad_raster
        okSgl = np.abs(Sgl) <= max_slew

        for ii in range(3):
            if okGopt[ii] == 1 and okSopt[ii] == 1:
                k_left[ii] = kopt[ii]
            elif okGsl[ii] == 1:
                k_left[ii] = ksl[ii]
            elif okSgl[ii] == 1:
                k_left[ii] = kgl[ii]
            else:
                print('Unknown error')

        success, k = __joinright1(k_left, k_end, (k_left - k0) / grad_raster, G_end, use_points - 1)
        if len(k) != 0:
            if len(k.shape) == 1:
                k = k.reshape((len(k), 1))
            if len(k_left.shape) == 1:
                k_left = k_left.reshape((len(k_left), 1))
            k_out_left = np.hstack((k_left, k))
        else:
            k_out_left = k_left

        return success, k_out_left

    def __joinright0(k0, k_end, G0, G_end, use_points):
        if use_points == 0:
            G = np.stack((G0, (k_end - k0) / grad_raster, G_end)).T
            S = (G[:, 1:] - G[:, :-1]) / grad_raster

            k_out_right = np.zeros((3, 0))
            success = __inside_limits(G, S)

            return success, k_out_right

        dk = (k0 - k_end) / (use_points + 1)
        kopt = k_end + dk
        Gopt = (k_end - kopt) / grad_raster
        Sopt = (G_end - Gopt) / grad_raster

        okGopt = np.sum(np.square(Gopt)) <= np.square(max_grad)
        okSopt = np.sum(np.square(Sopt)) <= np.square(max_slew)

        if okGopt and okSopt:
            k_right = kopt
        else:
            a = np.multiply(max_grad, grad_raster)
            b = np.multiply(max_slew, grad_raster ** 2)

            dkprol = -G_end * grad_raster
            dkconn = dk - dkprol

            ksl = k_end + dkprol + dkconn / np.linalg.norm(dkconn) * b
            Gsl = (k_end - ksl) / grad_raster
            okGsl = np.sum(np.square(Gsl)) <= np.square(max_grad)

            kgl = k_end + np.multiply(dk / np.linalg.norm(dk), a)
            Ggl = (k_end - kgl) / grad_raster
            Sgl = (G_end - Ggl) / grad_raster
            okSgl = np.sum(np.square(Sgl)) <= np.square(max_slew)

            if okGsl:
                k_right = ksl
            elif okSgl:
                k_right = kgl
            else:
                c = np.linalg.norm(dkprol)
                c1 = np.divide(np.square(a) - np.square(b) + np.square(c), (2 * c))
                h = np.sqrt(np.square(a) - np.square(c1))
                kglsl = k_end + np.multiply(c1, np.divide(dkprol, np.linalg.norm(dkprol)))
                projondkprol = (kgl * dkprol.T) * (dkprol / np.linalg.norm(dkprol))
                hdirection = kgl - projondkprol
                kglsl = kglsl + h * hdirection / np.linalg.norm(hdirection)
                k_right = kglsl

        success, k = __joinleft0(k0, k_right, G0, (k_end - k_right) / grad_raster, use_points - 1)
        if len(k) != 0:
            if len(k.shape) == 1:
                k = k.reshape((len(k), 1))
            if len(k_right.shape) == 1:
                k_right = k_right.reshape((len(k_right), 1))
            k_out_right = np.hstack((k, k_right))
        else:
            k_out_right = k_right

        return success, k_out_right

    def __joinright1(k0, k_end, G0, G_end, use_points):
        if use_points == 0:
            G = np.stack((G0, (k_end - k0) / grad_raster, G_end))
            S = (G[:, 1:] - G[:, :-1]) / grad_raster

            k_out_right = np.zeros((3, 0))
            success = __inside_limits(G, S)

            return success, k_out_right

        k_right = np.zeros(3)

        dk = (k0 - k_end) / (use_points + 1)
        kopt = k_end + dk
        Gopt = (k_end - kopt) / grad_raster
        Sopt = (G_end - Gopt) / grad_raster

        okGopt = np.abs(Gopt) <= max_grad
        okSopt = np.abs(Sopt) <= max_slew

        dkprol = -G_end * grad_raster
        dkconn = dk - dkprol

        ksl = k_end + dkprol + np.multiply(np.sign(dkconn), max_slew) * grad_raster ** 2
        Gsl = (k_end - ksl) / grad_raster
        okGsl = np.abs(Gsl) <= max_grad

        kgl = k_end + np.multiply(np.sign(dk), max_grad) * grad_raster
        Ggl = (k_end - kgl) / grad_raster
        Sgl = (G_end - Ggl) / grad_raster
        okSgl = np.abs(Sgl) <= max_slew

        for ii in range(3):
            if okGopt[ii] == 1 and okSopt[ii] == 1:
                k_right[ii] = kopt[ii]
            elif okGsl[ii] == 1:
                k_right[ii] = ksl[ii]
            elif okSgl[ii] == 1:
                k_right[ii] = kgl[ii]
            else:
                print('Unknown error')

        success, k = __joinleft1(k0, k_right, G0, (k_end - k_right) / grad_raster, use_points - 1)
        if len(k) != 0:
            if len(k.shape) == 1:
                k = k.reshape((len(k), 1))
            if len(k_right.shape) == 1:
                k_right = k_right.reshape((len(k_right), 1))
            k_out_right = np.hstack((k, k_right))
        else:
            k_out_right = k_right

        return success, k_out_right

    # =========
    # MAIN FUNCTION
    # =========
    if np.all(np.where(max_grad <= 0)):
        max_grad = [system.max_grad]
    if np.all(np.where(max_slew <= 0)):
        max_slew = [system.max_slew]

    grad_raster = system.grad_raster_time

    if len(max_grad) == 1 and len(max_slew) == 1:
        mode = 0
    elif len(max_grad) == 3 and len(max_slew) == 3:
        mode = 1
    else:
        raise ValueError('Input value max grad or max slew in invalid format.')

    G0 = (k0[:, 1] - k0[:, 0]) / grad_raster
    G_end = (k_end[:, 1] - k_end[:, 0]) / grad_raster
    k0 = k0[:, 1]
    k_end = k_end[:, 0]

    success = 0
    k_out = np.zeros((3, 0))
    use_points = 0

    while success == 0 and use_points <= max_points:
        if mode == 0:
            if np.linalg.norm(G0) > max_grad or np.linalg.norm(G_end) > max_grad:
                break
            success, k_out = __joinleft0(k0, k_end, G0, G_end, use_points)
        else:
            if np.abs(G0) > np.abs(max_grad) or np.abs(G_end) > np.abs(max_grad):
                break
            success, k_out = __joinleft1(k0, k_end, G0, G_end, use_points)
        use_points += 1

    return k_out, success



import math
from types import SimpleNamespace

import matplotlib as mpl
import numpy as np

mpl.use('TkAgg')
from matplotlib import pyplot as plt

from pypulseq.Sequence.test_report import test_report as ext_test_report
from pypulseq.check_timing import check_timing as ext_check_timing
from pypulseq.Sequence import block
from pypulseq.Sequence.read_seq import read
from pypulseq.Sequence.write_seq import write
from pypulseq.calc_duration import calc_duration
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.decompress_shape import decompress_shape
from pypulseq.event_lib import EventLibrary
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform


class Sequence:
    def __init__(self, system=Opts()):
        self.version_major = 1
        self.version_minor = 2
        self.version_revision = 0
        self.system = system
        self.definitions = dict()
        self.grad_library = EventLibrary()
        self.shape_library = EventLibrary()
        self.rf_library = EventLibrary()
        self.adc_library = EventLibrary()
        self.delay_library = EventLibrary()
        self.block_events = {}
        self.rf_raster_time = self.system.rf_raster_time
        self.grad_raster_time = self.system.grad_raster_time

    def __str__(self):
        s = "Sequence:"
        s += "\nshape_library: " + str(self.shape_library)
        s += "\nrf_library: " + str(self.rf_library)
        s += "\ngrad_library: " + str(self.grad_library)
        s += "\nadc_library: " + str(self.adc_library)
        s += "\ndelay_library: " + str(self.delay_library)
        s += "\nrf_raster_time: " + str(self.rf_raster_time)
        s += "\ngrad_raster_time: " + str(self.grad_raster_time)
        s += "\nblock_events: " + str(len(self.block_events))
        return s

    def duration(self):
        """
        Get duration of this sequence.

        Returns
        -------
        duration : float
            Duration of this sequence in millis.
        num_blocks : int
            Number of blocks in this sequence.
        event_count : int
            Number of events in this sequence.
        """

        num_blocks = len(self.block_events)
        event_count = np.zeros(len(self.block_events[1]))
        duration = 0
        for i in range(num_blocks):
            block = self.get_block(i + 1)
            event_count += self.block_events[i + 1] > 0
            duration += calc_duration(block)

        return duration, num_blocks, event_count

    def check_timing(self):
        """
        Check timing of the events in each block based on grad raster time system limit.

        Returns
        -------
        is_ok : bool
            Boolean flag indicating timing errors.
        error_report : str
            Error report in case of timing errors.
        """
        num_blocks = len(self.block_events)
        is_ok = True
        error_report = []
        for i in range(num_blocks):
            block = self.get_block(i + 1)
            event_names = ['rf', 'gx', 'gy', 'gz', 'adc', 'delay']
            ind = [hasattr(block, 'rf'), hasattr(block, 'gx'), hasattr(block, 'gy'), hasattr(block, 'gz'),
                   hasattr(block, 'adc'), hasattr(block, 'delay')]
            ev = [getattr(block, event_names[i]) for i in range(len(event_names)) if ind[i] == 1]
            res, rep = ext_check_timing(self, *ev)
            is_ok = is_ok and res
            if len(rep) != 0:
                error_report.append(f'Block: {i} - {rep}\n')

        return is_ok, error_report

    def test_report(self):
        """
        Analyze the sequence and return a text report.
        """
        return ext_test_report(self)

    def set_definition(self, key: str, val: str):
        """
        Sets custom definition to the `Sequence`.

        Parameters
        ----------
        key : str
            Definition key.
        val : str
            Definition value.
        """
        self.definitions[key] = val

    def get_definition(self, key: str) -> str:
        """
        Retrieves definition identified by `key` from `Sequence`. Returns `None` if no matching definition is found.

        Parameters
        ----------
        key : str
            Key of definition to retrieve.

        Returns
        -------
        str
            Definition identified by `key` if found, else returns `None`.
        """
        if key in self.definitions:
            return self.definitions[key]
        else:
            return None

    def add_block(self, *args):
        """
        Adds event(s) as a block to `Sequence`.

        Parameters
        ----------
        args
            Event or list of events to be added as a block to `Sequence`.
        """
        block.add_block(self, len(self.block_events) + 1, *args)

    def get_block(self, block_index: int) -> SimpleNamespace:
        """
        Retrieves block of events identified by `block_index` from `Sequence`.

        Parameters
        ----------
        block_index : int
            Index of block to be retrieved from `Sequence`.

        Returns
        -------
        SimpleNamespace
            Block identified by `block_index`.
        """
        return block.get_block(self, block_index)

    def read(self, file_path: str):
        """
        Read `.seq` file from `file_path`.

        Parameters
        ----------
        file_path : str
            Path to `.seq` file to be read.
        """
        read(self, file_path)

    def write(self, name: str):
        """
        Writes the calling `Sequence` object as a `.seq` file with filename `name`.

        Parameters
        ----------
        name :str
            Filename of `.seq` file to be written to disk.
        """
        write(self, name)

    def rf_from_lib_data(self, lib_data):
        """
        Construct RF object from `lib_data`.

        Parameters
        ----------
        lib_data : list
            RF envelope.

        Returns
        -------
        rf : SimpleNamespace
            RF object constructed from lib_data.
        """
        rf = SimpleNamespace()
        rf.type = 'rf'

        amplitude, mag_shape, phase_shape = lib_data[0], lib_data[1], lib_data[2]
        shape_data = self.shape_library.data[mag_shape]
        compressed = SimpleNamespace()
        compressed.num_samples = shape_data[0]
        compressed.data = shape_data[1:]
        mag = decompress_shape(compressed)
        shape_data = self.shape_library.data[phase_shape]
        compressed.num_samples = shape_data[0]
        compressed.data = shape_data[1:]
        phase = decompress_shape(compressed)
        rf.signal = 1j * 2 * np.pi * phase
        rf.signal = amplitude * mag * np.exp(rf.signal)
        rf.t = np.arange(1, max(mag.shape) + 1) * self.rf_raster_time

        if max(lib_data.shape) < 6:
            rf.delay = 0
            rf.freq_offset = lib_data[3]
            rf.phase_offset = lib_data[4]
            lib_data = np.append(lib_data, 0)
        else:
            rf.delay = lib_data[3]
            rf.freq_offset = lib_data[4]
            rf.phase_offset = lib_data[5]

        if max(lib_data.shape) < 7:
            lib_data = np.append(lib_data, 0)
        rf.dead_time = lib_data[6]

        if max(lib_data.shape) < 8:
            lib_data = np.append(lib_data, 0)
        rf.ringdown_time = lib_data[7]

        if max(lib_data.shape) < 9:
            lib_data = np.append(lib_data, 0)

        use_cases = {1: 'excitation', 2: 'refocusing', 3: 'inversion'}
        if lib_data[8] in use_cases:
            rf.use = use_cases[lib_data[8]]

        return rf

    def calculate_kspace(self, trajectory_delay: int = 0):
        """
        Calculates the k-space trajectory of the entire pulse sequence.

        Parameters
        ----------
        trajectory_delay : int
            Compensation factor in millis to align ADC and gradients in the reconstruction.

        Returns
        -------
        k_traj_adc : numpy.ndarray
            K-space trajectory sampled at `t_adc` timepoints.
        k_traj : numpy.ndarray
            K-space trajectory of the entire pulse sequence.
        t_excitation : numpy.ndarray
            Excitation timepoints.
        t_refocusing : numpy.ndarray
            Refocusing timepoints.
        t_adc : numpy.ndarray
            Sampling timepoints.
        """
        c_excitation = 0
        c_refocusing = 0
        c_adc_samples = 0

        for i in range(len(self.block_events)):
            block = self.get_block(i + 1)
            if hasattr(block, 'rf'):
                if not hasattr(block.rf, 'use') or block.rf.use != 'refocusing':
                    c_excitation += 1
                else:
                    c_refocusing += 1

            if hasattr(block, 'adc'):
                c_adc_samples += int(block.adc.num_samples)

        t_excitation = np.zeros(c_excitation)
        t_refocusing = np.zeros(c_refocusing)
        k_time = np.zeros(c_adc_samples)
        current_dur = 0
        c_excitation = 0
        c_refocusing = 0
        k_counter = 0
        traj_recon_delay = trajectory_delay

        for i in range(len(self.block_events)):
            block = self.get_block(i + 1)
            if hasattr(block, 'rf'):
                rf = block.rf
                rf_center, _ = calc_rf_center(rf)
                t = rf.delay + rf_center
                if not hasattr(block.rf, 'use') or block.rf.use != 'refocusing':
                    t_excitation[c_excitation] = current_dur + t
                    c_excitation += 1
                else:
                    t_refocusing[c_refocusing] = current_dur + t
                    c_refocusing += 1
            if hasattr(block, 'adc'):
                k_time[k_counter:k_counter + block.adc.num_samples] = np.arange(
                    block.adc.num_samples) * block.adc.dwell + block.adc.delay + current_dur + traj_recon_delay
                k_counter += block.adc.num_samples
            current_dur += calc_duration(block)

        gw = self.gradient_waveforms()
        i_excitation = np.round(t_excitation / self.grad_raster_time)
        i_refocusing = np.round(t_refocusing / self.grad_raster_time)
        k_traj = np.zeros(gw.shape)
        k = [0, 0, 0]
        for i in range(gw.shape[1]):
            k += gw[:, i] * self.grad_raster_time
            k_traj[:, i] = k
            if len(np.where(i_excitation == i + 1)[0]) >= 1:
                k = 0
                k_traj[:, i] = np.nan
            if len(np.where(i_refocusing == i + 1)[0]) >= 1:
                k = -k

        k_traj_adc = []
        for i in range(k_traj.shape[0]):
            k_traj_adc.append(np.interp(k_time, np.arange(1, k_traj.shape[1] + 1) * self.grad_raster_time, k_traj[i]))
        k_traj_adc = np.asarray(k_traj_adc)
        t_adc = k_time

        return k_traj_adc, k_traj, t_excitation, t_refocusing, t_adc

    def gradient_waveforms(self) -> np.ndarray:
        """
        Decompress the entire gradient waveform. Returns an array of shape `gradient_axesxtimepoints`.
        `gradient_axes` is typically 3.

        Returns
        -------
        grad_waveforms : numpy.ndarray
            Decompressed gradient waveform.
        """
        duration, num_blocks, _ = self.duration()

        wave_length = math.ceil(duration / self.grad_raster_time)
        grad_channels = 3
        grad_waveforms = np.zeros((grad_channels, wave_length))
        grad_channels = ['gx', 'gy', 'gz']

        t0 = 0
        t0_n = 0
        for i in range(num_blocks):
            block = self.get_block(i + 1)
            for j in range(len(grad_channels)):
                if hasattr(block, grad_channels[j]):
                    grad = getattr(block, grad_channels[j])
                    if grad.type == 'grad':
                        nt_start = round((grad.delay + grad.t[0]) / self.grad_raster_time)
                        waveform = grad.waveform
                    else:
                        nt_start = round(grad.delay / self.grad_raster_time)
                        if abs(grad.flat_time) > np.finfo(float).eps:
                            t = np.cumsum([0, grad.rise_time, grad.flat_time, grad.fall_time])
                            trap_form = np.multiply([0, 1, 1, 0], grad.amplitude)
                        else:
                            t = np.cumsum([0, grad.rise_time, grad.fall_time])
                            trap_form = np.multiply([0, 1, 0], grad.amplitude)

                        tn = math.floor(t[-1] / self.grad_raster_time)
                        t = np.append(t, t[-1] + self.grad_raster_time)
                        trap_form = np.append(trap_form, 0)

                        if abs(grad.amplitude) > np.finfo(float).eps:
                            waveform = points_to_waveform(t, trap_form, self.grad_raster_time)
                        else:
                            waveform = np.zeros(tn + 1)

                    if waveform.size != np.sum(np.isfinite(waveform)):
                        raise Warning('Not all elements of the generated waveform are finite')

                    """
                    Matlab dynamically resizes arrays during slice assignment operation if assignment is out of bounds
                    Numpy does not
                    Following lines are a workaround
                    """
                    l1, l2 = int(t0_n + nt_start), int(t0_n + nt_start + max(waveform.shape))
                    if l2 > grad_waveforms.shape[1]:
                        grad_waveforms.resize((len(grad_channels), l2))
                    grad_waveforms[j, l1:l2] = waveform

            t0 += calc_duration(block)
            t0_n = round(t0 / self.grad_raster_time)

        return grad_waveforms

    def plot(self, type: str = 'Gradient', time_range=(0, np.inf), time_disp: str = 's', save: bool=False):
        """
        Plot `Sequence`.

        Parameters
        ----------
        type : str
            Gradients display type, must be one of either 'Gradient' or 'Kspace'.
        time_range : List
            Time range (x-axis limits) for plotting the sequence. Default is 0 to infinity (entire sequence).
        time_disp : str
            Time display type, must be one of `s`, `ms` or `us`.
        save_as : bool
            Boolean flag indicating if plots should be saved. The two figures will be saved as JPG with numerical
            suffixes to the filename 'seq_plot'.
        """

        valid_plot_types = ['Gradient', 'Kspace']
        valid_time_units = ['s', 'ms', 'us']
        if type not in valid_plot_types:
            raise Exception()
        if time_disp not in valid_time_units:
            raise Exception()

        fig1, fig2 = plt.figure(1), plt.figure(2)
        sp11 = fig1.add_subplot(311)
        sp12, sp13 = fig1.add_subplot(312, sharex=sp11), fig1.add_subplot(313, sharex=sp11)
        fig2_sp_list = [fig2.add_subplot(311, sharex=sp11), fig2.add_subplot(312, sharex=sp11),
                        fig2.add_subplot(313, sharex=sp11)]

        t_factor_list = [1, 1e3, 1e6]
        t_factor = t_factor_list[valid_time_units.index(time_disp)]
        t0 = 0
        for iB in range(1, len(self.block_events) + 1):
            block = self.get_block(iB)
            is_valid = time_range[0] <= t0 <= time_range[1]
            if is_valid:
                if hasattr(block, 'adc'):
                    adc = block.adc
                    t = adc.delay + [(x * adc.dwell) for x in range(0, int(adc.num_samples))]
                    sp11.plot((t0 + t), np.zeros(len(t)), 'rx')
                if hasattr(block, 'rf'):
                    rf = block.rf
                    tc, ic = calc_rf_center(rf)
                    t = rf.t + rf.delay
                    tc = tc + rf.delay
                    sp12.plot(t_factor * (t0 + t), abs(rf.signal))
                    sp13.plot(t_factor * (t0 + t), np.angle(
                        rf.signal * np.exp(1j * rf.phase_offset) * np.exp(1j * 2 * math.pi * rf.t * rf.freq_offset)),
                              t_factor * (t0 + tc), np.angle(rf.signal[ic] * np.exp(1j * rf.phase_offset) * np.exp(
                            1j * 2 * math.pi * rf.t[ic] * rf.freq_offset)), 'xb')
                grad_channels = ['gx', 'gy', 'gz']
                for x in range(0, len(grad_channels)):
                    if hasattr(block, grad_channels[x]):
                        grad = getattr(block, grad_channels[x])
                        if grad.type == 'grad':
                            # In place unpacking of grad.t with the starred expression
                            t = grad.delay + [0, *(grad.t + (grad.t[1] - grad.t[0]) / 2),
                                              grad.t[-1] + grad.t[1] - grad.t[0]]
                            waveform = np.array([grad.first, grad.last])
                            waveform = 1e-3 * np.insert(waveform, 1, grad.waveform)
                        else:
                            t = np.cumsum([0, grad.delay, grad.rise_time, grad.flat_time, grad.fall_time])
                            waveform = 1e-3 * grad.amplitude * np.array([0, 0, 1, 1, 0])
                        fig2_sp_list[x].plot(t_factor * (t0 + t), waveform)
            t0 += calc_duration(block)

        grad_plot_labels = ['x', 'y', 'z']
        sp11.set_ylabel('ADC')
        sp12.set_ylabel('RF mag (Hz)')
        sp13.set_ylabel('RF phase (rad)')
        [fig2_sp_list[x].set_ylabel(f'G{grad_plot_labels[x]} (kHz/m)') for x in range(3)]
        # Setting display limits
        disp_range = t_factor * np.array([time_range[0], min(t0, time_range[1])])
        sp11.set_xlim(disp_range)
        sp12.set_xlim(disp_range)
        sp13.set_xlim(disp_range)
        [x.set_xlim(disp_range) for x in fig2_sp_list]

        fig1.tight_layout()
        fig2.tight_layout()
        if save:
            fig1.savefig('seq_plot1.jpg')
            fig2.savefig('seq_plot2.jpg')
        plt.show()



import math
from types import SimpleNamespace

import numpy as np

from pypulseq.make_trap_pulse import make_trapezoid
from pypulseq.opts import Opts


def make_arbitrary_rf(signal: np.ndarray, flip_angle: float, system: Opts = Opts(), freq_offset: float = 0,
                      phase_offset: float = 0, time_bw_product: float = 0, bandwidth: float = 0, max_grad: float = 0,
                      max_slew: float = 0, slice_thickness: float = 0, delay: float = 0,
                      use: str = None) -> SimpleNamespace:
    """
    Creates a radio-frequency pulse event with arbitrary pulse shape and optionally an accompanying slice select
    trapezoidal gradient event.

    Parameters
    ----------
    signal : np.ndarray
        Arbitrary waveform.
    flip_angle : float
        Flip angle in radians.
    system : Opts
        System limits. Default is a system limits object initialised to default values.
    freq_offset : float
        Frequency offset in Hertz (Hz). Default is 0.
    phase_offset : float
        Phase offset in Hertz (Hz). Default is 0.
    time_bw_product : float
        Time-bandwidth product. Default is 4.
    bandwidth : float
        Bandwidth in Hertz (Hz).
     max_grad : float
        Maximum gradient strength of accompanying slice select trapezoidal event. Default is `system`'s `max_grad`.
     max_slew : float
        Maximum slew rate of accompanying slice select trapezoidal event. Default is `system`'s `max_slew`.
    slice_thickness : float
        Slice thickness of accompanying slice select trapezoidal event. The slice thickness determines the area of the
        slice select event.
    delay : float
        Delay in milliseconds (ms) of accompanying slice select trapezoidal event.
    use : str
        Use of arbitrary radio-frequency pulse event. Must be one of 'excitation', 'refocusing' or 'inversion'.

    Returns
    -------
    rf : SimpleNamespace
        Radio-frequency pulse event with arbitrary pulse shape.
    gz : SimpleNamespace
        Slice select trapezoidal gradient event accompanying the arbitrary radio-frequency pulse event.
    """
    valid_use_pulses = ['excitation', 'refocusing', 'inversion']
    if use is not None and use not in valid_use_pulses:
        raise ValueError(
            f'Invalid use parameter. Must be one of excitation, refocusing or inversion. You passed: {use}')

    signal = np.squeeze(signal)
    if signal.ndim > 1:
        raise ValueError(f'signal should have ndim=1. You passed ndim={signal.ndim}')
    signal = signal / np.sum(signal * system.rf_raster_time) * flip_angle / (2 * np.pi)

    N = len(signal)
    duration = N * system.rf_raster_time
    t = np.arange(1, N + 1) * system.rf_raster_time

    rf = SimpleNamespace()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.dead_time = system.rf_dead_time
    rf.ringdown_time = system.rf_ringdown_time
    rf.delay = delay

    if use is not None:
        rf.use = use

    if rf.dead_time > rf.delay:
        rf.delay = rf.dead_time

    try:
        if slice_thickness <= 0:
            raise ValueError('Slice thickness must be provided.')
        if bandwidth <= 0:
            raise ValueError('Bandwidth of pulse must be provided.')

        if max_grad > 0:
            system.max_grad = max_grad
        if max_slew > 0:
            system.max_slew = max_slew

        BW = bandwidth
        if time_bw_product > 0:
            BW = time_bw_product / duration

        amplitude = BW / slice_thickness
        area = amplitude * duration
        gz = make_trapezoid(channel='z', system=system, flat_time=duration, flat_area=area)

        if rf.delay > gz.rise_time:
            gz.delay = math.ceil((rf.delay - gz.rise_time) / system.grad_raster_time) * system.grad_raster_time

        if rf.delay < (gz.rise_time + gz.delay):
            rf.delay = gz.rise_time + gz.delay
    except:
        gz = None

    if rf.ringdown_time > 0:
        t_fill = np.arange(1, round(rf.ringdown_time / 1e-6) + 1) * 1e-6
        rf.t = np.concatenate((rf.t, rf.t[-1] + t_fill))
        rf.signal = np.concatenate((rf.signal, np.zeros(len(t_fill))))

    return rf, gz



import numpy as np

from pypulseq.calc_duration import calc_duration


def align(*args) -> list:
    """
    Aligns `SimpleNamespace` blocks as per specified alignment options by setting delays of the pulse sequence events
    within the block. All previously configured delays within objects are taken into account during calculating of the
    block duration but then reset according to the selected alignment. Possible values for align_spec are 'left',
    'center', 'right'.

    Parameters
    ----------
    args : list
        List of alignment options and `SimpleNamespace` blocks.
        Template: [alignment_spec, 'SimpleNamespace` block, [alignment_spec, `SimpleNamespace` block, ...]].
        Alignment spec can be one of `left`, `center` or `right`.

    Returns
    -------
    objects : list
        List of aligned `SimpleNamespace` blocks.
    """
    alignment_options = ['left', 'center', 'right']
    if not isinstance(args[0], str):
        raise ValueError('First parameter must be of type str.')

    curr_align = alignment_options.index(args[0]) if args[0] in alignment_options else None

    i_objects = []
    alignments = []
    for i in range(1, len(args)):
        if curr_align is None:
            raise ValueError('Invalid alignment spec.')
        if isinstance(args[i], str):
            curr_align = alignment_options.index(args[i]) if args[i] in alignment_options else None
            continue
        i_objects.append(i)
        alignments.append(curr_align)

    args = np.array(args)
    objects = args[i_objects]
    dur = calc_duration(*objects)

    for i in range(len(objects)):
        if alignments[i] == 0:
            objects[i].delay = 0
        elif alignments[i] == 1:
            objects[i].delay = (dur - calc_duration(objects[i])) / 2
        elif alignments[i] == 2:
            objects[i].delay = dur - calc_duration(objects[i]) + objects[i].delay

    return objects



from pathlib import Path

import numpy as np

from pypulseq.event_lib import EventLibrary

def read(self, path: str, **kwargs):
    """
    Reads a `.seq` file from `path`.

    Parameters
    ----------
    path : Path
        Path of .seq file to be read.
    """

    detect_rf_use = True if 'detect_rf_use' in kwargs else False

    input_file = open(path, 'r')
    self.shape_library = EventLibrary()
    self.rf_library = EventLibrary()
    self.grad_library = EventLibrary()
    self.adc_library = EventLibrary()
    self.delay_library = EventLibrary()
    self.block_events = {}
    self.rf_raster_time = self.system.rf_raster_time
    self.grad_raster_time = self.system.grad_raster_time
    self.definitions = {}

    jemris_generated = False

    while True:
        section = __skip_comments(input_file)
        if section == -1:
            break

        if section == '[DEFINITIONS]':
            self.definitions = __read_definitions(input_file)
        elif section == '[JEMRIS]': # Added by GT
            jemris_generated = True
        elif section == '[VERSION]':
            version_major, version_minor, version_revision = __read_version(input_file)
            if version_major != self.version_major or version_minor != self.version_minor or version_revision != self.version_revision:
                raise Exception(
                    f'Unsupported version: {version_major, version_minor, version_revision}. '
                    f'Expected: {self.version_major, self.version_minor, self.version_revision}')
        elif section == '[BLOCKS]':
            self.block_events = __read_blocks(input_file)
        elif section == '[RF]':
            if jemris_generated:
                self.rf_library = __read_events(input_file, [1, 1, 1, 1, 1])
            else:
                self.rf_library = __read_events(input_file, [1, 1, 1, 1e-6, 1, 1])
        elif section == '[GRADIENTS]':
            self.grad_library = __read_events(input_file, [1, 1, 1e-6], 'g', self.grad_library)
        elif section == '[TRAP]':
            if jemris_generated:
                self.grad_library = __read_events(input_file, [1, 1e-6, 1e-6, 1e-6], 't', self.grad_library)
            else:
                self.grad_library = __read_events(input_file, [1, 1e-6, 1e-6, 1e-6, 1e-6], 't', self.grad_library)
        elif section == '[ADC]':
            self.adc_library = __read_events(input_file, [1, 1e-9, 1e-6, 1, 1])
        elif section == '[DELAYS]':
            self.delay_library = __read_events(input_file, 1e-6)
        elif section == '[SHAPES]':
            self.shape_library = __read_shapes(input_file)
        else:
            raise ValueError(f'Unknown section code: {section}')

    if detect_rf_use:
        for k in self.rf_library.keys():
            lib_data = self.rf_library.data[k]
            rf = self.rf_from_lib_data(lib_data)
            flip_deg = abs(np.sum(rf.signal)) * rf.t[0] * 360
            if len(lib_data) < 9:
                if flip_deg <= 90:
                    lib_data[8] = 0
                else:
                    lib_data[8] = 2
                self.rf_library.data[k] = lib_data



def __read_definitions(input_file) -> dict:
    """
    Read definitions from .seq file.

    Parameters
    ----------
    input_file : file object
        .seq file

    Returns
    -------
    definitions : dict
        Dict object containing key value pairs of definitions.
    """
    definitions = dict()
    line = __strip_line(input_file)
    while line != '' and line[0] != '#':
        tok = line.split(' ')
        try:  # Try converting every element into a float
            [float(x) for x in tok[1:]]
            definitions[tok[0]] = np.array(tok[1:], dtype=float)
        except ValueError:  # Try clause did not work!
            definitions[tok[0]] = tok[1:]
        line = __strip_line(input_file)

    return definitions


def __read_version(input_file) -> tuple:
    """
    Read version from .seq file.

    Parameters
    ----------
    input_file : file object
        .seq file

    Returns
    -------
    tuple
        Tuple of major, minor and revision number.
    """
    line = __strip_line(input_file)
    while line != '' and line[0] != '#':
        tok = line.split(' ')
        if tok[0] == 'major':
            major = np.array(tok[1:], dtype=float)
        elif tok[0] == 'minor':
            minor = np.array(tok[1:], dtype=float)
        elif tok[0] == 'revision':
            revision = np.array(tok[1:], dtype=float)
        line = __strip_line(input_file)

    return major, minor, revision


def __read_blocks(input_file) -> dict:
    """
    Read Pulseq blocks from .seq file.

    Parameters
    ----------
    input_file : file
        .seq file

    Returns
    -------
    event_table : dict
        Dict object containing key value pairs of Pulseq block ID and block definition.
    """

    line = __strip_line(input_file)

    event_table = dict()
    while line != '' and line != '#':
        block_events = np.fromstring(line, dtype=int, sep=' ')
        event_table[block_events[0]] = block_events[1:]
        line = __strip_line(input_file)

    return event_table


def __read_events(input_file, scale, type: str = None, event_library: EventLibrary = None) -> EventLibrary:
    """
    Read Pulseq events from .seq file.

    Parameters
    ----------
    input_file : file object
        .seq file
    scale : int
        Scaling factor.
    type : str
        Type of Pulseq event.
    event_library : EventLibrary
        EventLibrary

    Returns
    -------
    definitions : dict
        `EventLibrary` object containing Pulseq event definitions.
    """
    scale = 1 if scale is None else scale
    event_library = event_library if event_library is not None else EventLibrary()

    line = __strip_line(input_file)

    while line != '' and line != '#':
        data = np.fromstring(line, dtype=float, sep=' ')
        id = data[0]
        data = np.round(data[1:] * scale, decimals=6)
        if type is None:
            event_library.insert(key_id=id, new_data=data)
        else:
            event_library.insert(key_id=id, new_data=data, data_type=type)
        line = __strip_line(input_file)

    return event_library


def __read_shapes(input_file) -> EventLibrary:
    """
    Read Pulseq shapes from .seq file.

    Parameters
    ----------
    input_file : file
        .seq file

    Returns
    -------
    shape_library : EventLibrary
        `EventLibrary` object containing shape definitions.
    """
    shape_library = EventLibrary()

    line = __skip_comments(input_file)

    while line != -1 and (line != '' or line[0:8] == 'shape_id'):
        tok = line.split(' ')
        id = int(tok[1])
        line = __skip_comments(input_file)
        tok = line.split(' ')
        num_samples = int(tok[1])
        data = []
        line = __skip_comments(input_file)
        while line != '' and line != '#':
            data.append(float(line))
            line = __strip_line(input_file)
        line = __skip_comments(input_file)
        data.insert(0, num_samples)
        data = np.asarray(data)
        shape_library.insert(key_id=id, new_data=data)
    return shape_library


def __skip_comments(input_file) -> str:
    """
    Skip one '#' comment in .seq file.

    Parameters
    ----------
    input_file : file
        .seq file

    Returns
    -------
    line : str
        First line in `input_file` after skipping one '#' comment block. Note: File pointer is remembered, so successive calls work as expected.
    """

    line = __strip_line(input_file)

    while line != -1 and (line == '' or line[0] == '#'):
        line = __strip_line(input_file)

    return line


def __strip_line(input_file):
    """
    Removes spaces and newline whitespaces.

    Parameters
    ----------
    input_file : file
        .seq file

    Returns
    -------
    line : str
        First line in input_file after removing spaces and newline whitespaces. Note: File pointer is remembered,
        so successive calls work as expected. Returns -1 for eof.
    """
    line = input_file.readline()  # If line is an empty string, end of the file has been reached
    return line.strip() if line != '' else -1



from types import SimpleNamespace

import numpy as np

from pypulseq.calc_duration import calc_duration
from pypulseq.make_arbitrary_grad import make_arbitrary_grad
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform


def add_gradients(grads: list, system=Opts(), max_grad: int = 0, max_slew: int = 0) -> np.ndarray:
    """
    Superpose several gradient events.

    Parameters
    ----------
    grads : list
        List of 'SimpleNamespace' gradient events.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    max_grad : float, optional
        Maximum gradient amplitude. Default is 0.
    max_slew : float, optional
        Maximum slew rate. Default is 0.

    Returns
    -------
    grad : SimpleNamespace
        Superimposition of gradient events from `grads`.
    """
    max_grad = max_grad if max_grad > 0 else system.max_grad
    max_slew = max_slew if max_slew > 0 else system.max_slew

    if len(grads) < 2:
        raise Exception()

    # First gradient defines channel
    channel = grads[0].channel

    # Find out the general delay of all gradients and other statistics
    delays, firsts, lasts, durs = [], [], [], []
    for ii in range(len(grads)):
        delays.append(grads[ii].delay)
        firsts.append(grads[ii].first)
        lasts.append(grads[ii].last)
        durs.append(calc_duration(grads[ii]))

    # Convert to numpy.ndarray for fancy-indexing later on
    firsts, lasts = np.array(firsts), np.array(lasts)

    common_delay = min(delays)
    total_duration = max(durs)

    waveforms = dict()
    max_length = 0
    for ii in range(len(grads)):
        g = grads[ii]
        if g.type == 'grad':
            waveforms[ii] = g.waveform
        elif g.type == 'trap':
            if g.flat_time > 0:  # Triangle or trapezoid
                times = [g.delay - common_delay,
                         g.delay - common_delay + g.rise_time,
                         g.delay - common_delay + g.rise_time + g.flat_time,
                         g.delay - common_delay + g.rise_time + g.flat_time + g.fall_time]
                amplitudes = [0, g.amplitude, g.amplitude, 0]
            else:
                times = [g.delay - common_delay,
                         g.delay - common_delay + g.rise_time,
                         g.delay - common_delay + g.rise_time + g.flat_time]
                amplitudes = [0, g.amplitude, 0]
            waveforms[ii] = points_to_waveform(times=times, amplitudes=amplitudes,
                                               grad_raster_time=system.grad_raster_time)
        else:
            raise ValueError('Unknown gradient type')

        if g.delay - common_delay > 0:
            # Stop for numpy.arange is not g.delay - common_delay - system.grad_raster_time like in Matlab
            # so as to include the endpoint
            t_delay = np.arange(0, g.delay - common_delay, step=system.grad_raster_time)
            waveforms[ii] = np.insert(waveforms[ii], 0, t_delay)

        num_points = len(waveforms[ii])
        max_length = num_points if num_points > max_length else max_length

    w = np.zeros(max_length)
    for ii in range(len(grads)):
        wt = np.zeros(max_length)
        wt[0:len(waveforms[ii])] = waveforms[ii]
        w += wt

    grad = make_arbitrary_grad(channel, w, system, max_slew=max_slew, max_grad=max_grad, delay=common_delay)
    grad.first = np.sum(firsts[np.where(delays == common_delay)])
    grad.last = np.sum(lasts[np.where(durs == total_duration)])

    return grad



import math
from types import SimpleNamespace

import numpy as np

from pypulseq.make_trap_pulse import make_trapezoid
from pypulseq.opts import Opts


def make_sinc_pulse(flip_angle: float, system: Opts = Opts(), duration: float = 0, freq_offset: float = 0,
                    phase_offset: float = 0, time_bw_product: float = 4, apodization: float = 0,
                    center_pos: float = 0.5, max_grad: float = 0, max_slew: float = 0, slice_thickness: float = 0,
                    delay: float = 0, use: str = None):
    """
    Creates a radio-frequency sinc pulse event and optionally accompanying slice select and slice select rephasing
    trapezoidal gradient events.

    Parameters
    ----------
    flip_angle : float
        Flip angle in radians.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    duration : float, optional
        Duration in milliseconds (ms). Default is 0.
    freq_offset : float, optional
        Frequency offset in Hertz (Hz). Default is 0.
    phase_offset : float, optional
        Phase offset in Hertz (Hz). Default is 0.
    time_bw_product : float, optional
        Time-bandwidth product. Default is 4.
    apodization : float, optional
        Apodization. Default is 0.
    center_pos : float, optional
        Position of peak. Default is 0.5 (midway).
     max_grad : float, optional
        Maximum gradient strength of accompanying slice select trapezoidal event. Default is 0.
     max_slew : float, optional
        Maximum slew rate of accompanying slice select trapezoidal event. Default is 0.
    slice_thickness : float, optional
        Slice thickness of accompanying slice select trapezoidal event. The slice thickness determines the area of the
        slice select event. Default is 0.
    delay : float, optional
        Delay in milliseconds (ms). Default is 0.
    use : str, optional
        Use of radio-frequency sinc pulse. Must be one of 'excitation', 'refocusing' or 'inversion'.

    Returns
    -------
    rf : SimpleNamespace
        Radio-frequency sinc pulse event.
    gz : SimpleNamespace, optional
        Accompanying slice select trapezoidal gradient event. Returned only if `slice_thickness` is provided.
    gzr : SimpleNamespace, optional
        Accompanying slice select rephasing trapezoidal gradient event. Returned only if `slice_thickness` is provided.
    """
    valid_use_pulses = ['excitation', 'refocusing', 'inversion']
    if use is not None and use not in valid_use_pulses:
        raise ValueError(
            f'Invalid use parameter. Must be one of excitation, refocusing or inversion. You passed: {use}')

    BW = time_bw_product / duration
    alpha = apodization
    N = int(round(duration / 1e-6))
    t = np.arange(1, N + 1) * system.rf_raster_time
    tt = t - (duration * center_pos)
    window = 1 - alpha + alpha * np.cos(2 * np.pi * tt / duration)
    signal = np.multiply(window, np.sinc(BW * tt))
    flip = np.sum(signal) * system.rf_raster_time * 2 * np.pi
    signal = signal * flip_angle / flip

    rf = SimpleNamespace()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.dead_time = system.rf_dead_time
    rf.ringdown_time = system.rf_ringdown_time
    rf.delay = delay
    if use is not None:
        rf.use = use

    if rf.dead_time > rf.delay:
        rf.delay = rf.dead_time

    try:
        if slice_thickness == 0:
            raise ValueError('Slice thickness must be provided')

        if max_grad > 0:
            system.max_grad = max_grad

        if max_slew > 0:
            system.max_slew = max_slew

        amplitude = BW / slice_thickness
        area = amplitude * duration
        gz = make_trapezoid(channel='z', system=system, flat_time=duration, flat_area=area)
        gzr = make_trapezoid(channel='z', system=system, area=-area * (1 - center_pos) - 0.5 * (gz.area - area))

        if rf.delay > gz.rise_time:
            gz.delay = math.ceil((rf.delay - gz.rise_time) / system.grad_raster_time) * system.grad_raster_time

        if rf.delay < (gz.rise_time + gz.delay):
            rf.delay = gz.rise_time + gz.delay
    except:
        gz = None
        gzr = None

    if rf.ringdown_time > 0:
        t_fill = np.arange(1, round(rf.ringdown_time / 1e-6) + 1) * 1e-6
        rf.t = np.concatenate((rf.t, rf.t[-1] + t_fill))
        rf.signal = np.concatenate((rf.signal, np.zeros(len(t_fill))))

    # Following 2 lines of code are workarounds for numpy returning 3.14... for np.angle(-0.00...)
    negative_zero_indices = np.where(rf.signal == -0.0)
    rf.signal[negative_zero_indices] = 0

    return rf, gz, gzr



import numpy as np


def k2g(k: np.ndarray, dt: float = 10e-6):
    """
    Derives gradient-waveforms from k-space trajectory in compliance with the dwell time.

    Parameters
    ----------
    k : numpy.ndarray
        Target k-space trajectory.
    dt : float
        Dwell time in milliseconds (ms). Default is 10e-6.
    Returns
    -------
    Gx, Gy, Gz  :numpy.ndarray
        Gradient waveforms to accomplish `k`.
    """
    factor = 1
    Gx = (factor / dt) * (np.diff(np.squeeze(k[0, :])))
    Gy = (factor / dt) * (np.diff(np.squeeze(k[1, :])))
    Gz = (factor / dt) * (np.diff(np.squeeze(k[2, :]))) if np.size(k, 0) > 2 else 0
    Gx = np.insert(Gx, 0, 0)
    Gy = np.insert(Gy, 0, 0)
    Gz = np.insert(Gz, 0, 0)
    return Gx, Gy, Gz



import numpy as np

from pypulseq.calc_ramp import calc_ramp
from pypulseq.opts import Opts
from copy import copy
from types import SimpleNamespace


def add_ramps(k, system=Opts(), rf: SimpleNamespace = None, max_grad: int = 0, max_slew: int = 0) -> list:
    """
    Adds segment so that `k` k-space trajectory ramps up from 0 to `k[0]` and ramps down from `k[-1]` to 0. If `k` is a
    tuple or list of k-space trajectories, ramp ups and ramp downs are added to each.

    Parameters
    ----------
    k : tuple, list or numpy.ndarray
        Tuple, list or numpy.ndarray of k-space trajectories to add ramp ups and ramp downs to.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    rf : SimpleNamespace, optional
        Zeros are added to this pulse sequence event over the ramp times in `k`.
    max_grad : int, optional
        Maximum gradient amplitude. Default is 0.
    max_slew : int, optional
        Maximum slew rate. Default is 0.

    Returns
    -------
    result : list
        List of ramped up and ramped down k-space trajectories from `k`.
    """
    if not isinstance(k, tuple) and not isinstance(k, list) and not isinstance(k, np.ndarray):
        raise Exception()

    k_arg = copy(k)
    if max_grad > 0:
        system.max_grad = max_grad

    if max_slew > 0:
        system.max_slew = max_slew

    k = np.vstack(k)
    num_channels = k.shape[0]
    k = np.vstack((k, np.zeros((3 - num_channels, k.shape[1]))))

    k_up, ok1 = calc_ramp(np.zeros((3, 2)), k[:, :2], system)
    k_down, ok2 = calc_ramp(k[:, -2:], np.zeros((3, 2)), system)
    if not (ok1 and ok2):
        raise Exception('Failed to calculate gradient ramps')

    k_up = np.hstack((np.zeros((3, 2)), k_up))
    k_down = np.hstack((k_down, np.zeros((3, 1))))

    k = np.hstack((k_up, k, k_down))

    result = []
    if not isinstance(k_arg, list):
        result.append(k[:num_channels])
    else:
        for i in range(num_channels):
            result.append(k[i])

    if rf is not None:
        result.append(np.concatenate((np.zeros(k_up.shape[1] * 10), rf, np.zeros(k_down.shape[1] * 10))))

    return result



from types import SimpleNamespace

from pypulseq.opts import Opts


def make_adc(num_samples: int = 0, system: Opts = Opts(), dwell: float = 0, duration: float = 0, delay: float = 0,
             freq_offset: float = 0, phase_offset: float = 0) -> SimpleNamespace:
    """
    Creates an ADC readout event.

    Parameters
    ----------
    num_samples: int, optional
        Number of readout samples.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    dwell : float, optional
        ADC dead time in milliseconds (ms) after sampling.
    duration : float, optional
        Duration in milliseconds (ms) of ADC readout event with `num_samples` number of samples.
    delay : float, optional
        Delay in milliseconds (ms) of ADC readout event.
    freq_offset : float, optional
        Frequency offset of ADC readout event.
    phase_offset : float, optional
        Phase offset of ADC readout event.

    Returns
    -------
    adc : SimpleNamespace
        ADC readout event.
    """
    adc = SimpleNamespace()
    adc.type = 'adc'
    adc.num_samples = num_samples
    adc.dwell = dwell
    adc.delay = delay
    adc.freq_offset = freq_offset
    adc.phase_offset = phase_offset
    adc.dead_time = system.adc_dead_time

    if (dwell == 0 and duration == 0) or (dwell > 0 and duration > 0):
        raise ValueError("Either dwell or duration must be defined")

    if duration > 0:
        adc.dwell = duration / num_samples

    if dwell > 0:
        adc.duration = dwell * num_samples

    return adc



import numpy as np

from pypulseq.calc_duration import calc_duration
from pypulseq.opts import Opts
from types import SimpleNamespace


def check_timing(system: Opts, *events: SimpleNamespace):
    """
    Checks if timings of events `events` are aligned with gradient raster time `system.grad_raster_time`.

    Parameters
    ----------
    system : Opts
        System limits object.
    events : SimpleNamespace
        Pulse events.

    Returns
    -------
    is_ok : bool
        Boolean flag indicating if timing of events `events` are aligned with gradient raster time
        `system.grad_raster_time`.
    text_err : str
        Error string, if timings are not aligned.
    """
    total_dur = calc_duration(*events)
    is_ok = __div_check(total_dur, system.grad_raster_time)
    if is_ok:
        text_err = str()
    else:
        text_err = f'Total duration: {total_dur * 1e6} us'

    for i in range(len(events)):
        e = events[i]
        ok = True
        if hasattr(e, 'type') and e.type == 'adc' or e.type == 'rf':
            raster = system.rf_raster_time
        else:
            raster = system.grad_raster_time

        if hasattr(e, 'delay'):
            if not __div_check(e.delay, raster):
                ok = False

        if hasattr(e, 'type') and e.type == 'trap':
            if not __div_check(e.rise_time, system.grad_raster_time) or \
                    not __div_check(e.flat_time, system.grad_raster_time) or \
                    not __div_check(e.fall_time, system.grad_raster_time):
                ok = False

        if not ok:
            is_ok = False
            if len(text_err) != 0:
                text_err += ' '

            text_err += '[ '
            if hasattr(e, 'type'):
                text_err += f'type: {e.type} '
            if hasattr(e, 'delay'):
                text_err += f'delay: {e.delay * 1e6} us '
            if hasattr(e, 'type') and e.type == 'trap':
                text_err += f'rise time: {e.rise_time * 1e6} flat time: {e.flat_time * 1e6} ' \
                            f'fall time: {e.fall_time * 1e6} us '
            text_err += ']'

    return is_ok, text_err


def __div_check(a, b):
    c = a / b
    return abs(c - np.round(c)) < 1e-9



from types import SimpleNamespace

import numpy as np

from pypulseq.calc_duration import calc_duration
from pypulseq.compress_shape import compress_shape
from pypulseq.decompress_shape import decompress_shape


def add_block(self, block_index: int, *args):
    """
    Inserts pulse sequence events into `self.block_events` at position `block_index`. Also performs gradient checks.

    Parameters
    ----------
    block_index : int
        Index at which `SimpleNamespace` events have to be inserted into `self.block_events`.
    args : list
        List of `SimpleNamespace` pulse sequence events to be added to `self.block_events`.
    """

    block_duration = calc_duration(*args)
    self.block_events[block_index] = np.zeros(6, dtype=np.int)
    duration = 0

    check_g = {}

    for event in args:
        if event.type == 'rf':
            mag = np.abs(event.signal)
            amplitude = np.max(mag)
            mag = np.divide(mag, amplitude)
            # Following line of code is a workaround for numpy's divide functions returning NaN when mathematical
            # edge cases are encountered (eg. divide by 0)
            mag[np.isnan(mag)] = 0
            phase = np.angle(event.signal)
            phase[phase < 0] += 2 * np.pi
            phase /= 2 * np.pi

            mag_shape = compress_shape(mag)
            data = np.insert(mag_shape.data, 0, mag_shape.num_samples)
            mag_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(mag_id, data)

            phase_shape = compress_shape(phase)
            data = np.insert(phase_shape.data, 0, phase_shape.num_samples)
            phase_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(phase_id, data)

            use = 0
            use_cases = {'excitation': 1, 'refocusing': 2, 'inversion': 3}
            if hasattr(event, 'use'):
                use = use_cases[event.use]

            data = np.array((amplitude, mag_id, phase_id, event.delay, event.freq_offset, event.phase_offset,
                             event.dead_time, event.ringdown_time, use))
            data_id, found = self.rf_library.find(data)
            if not found:
                self.rf_library.insert(data_id, data)

            self.block_events[block_index][1] = data_id
            duration = max(duration,
                           max(
                               mag.shape) * self.rf_raster_time + event.dead_time + event.ringdown_time + event.delay)
        elif event.type == 'grad':
            channel_num = ['x', 'y', 'z'].index(event.channel)
            idx = 2 + channel_num

            check_g[channel_num] = SimpleNamespace()
            check_g[channel_num].idx = idx
            check_g[channel_num].start = np.array((event.delay + min(event.t), event.first))
            check_g[channel_num].stop = np.array(
                (event.delay + max(event.t) + self.system.grad_raster_time, event.last))

            amplitude = max(abs(event.waveform))
            if amplitude > 0:
                g = event.waveform / amplitude
            else:
                g = event.waveform
            shape = compress_shape(g)
            data = np.insert(shape.data, 0, shape.num_samples)
            shape_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(shape_id, data)
            data = np.array([amplitude, shape_id, event.delay, event.first, event.last])
            grad_id, found = self.grad_library.find(data)
            if not found:
                self.grad_library.insert(grad_id, data, 'g')
            idx = 2 + channel_num
            self.block_events[block_index][idx] = grad_id
            duration = max(duration, len(g) * self.grad_raster_time)
        elif event.type == 'trap':
            channel_num = ['x', 'y', 'z'].index(event.channel)
            idx = 2 + channel_num
            check_g[channel_num] = SimpleNamespace()
            check_g[channel_num].idx = idx
            check_g[channel_num].start = np.array((0, 0))
            check_g[channel_num].stop = np.array((event.delay + event.rise_time + event.fall_time + event.flat_time, 0))
            data = np.array([event.amplitude, event.rise_time, event.flat_time, event.fall_time, event.delay])
            trap_id, found = self.grad_library.find(data)
            if not found:
                self.grad_library.insert(trap_id, data, 't')
            self.block_events[block_index][idx] = trap_id
            duration = max(duration, event.delay + event.rise_time + event.flat_time + event.fall_time)
        elif event.type == 'adc':
            data = np.array(
                [event.num_samples, event.dwell, max(event.delay, event.dead_time), event.freq_offset,
                 event.phase_offset, event.dead_time])
            adc_id, found = self.adc_library.find(data)
            if not found:
                self.adc_library.insert(adc_id, data)
            self.block_events[block_index][5] = adc_id
            duration = max(duration, event.delay + event.num_samples * event.dwell + event.dead_time)
        elif event.type == 'delay':
            data = np.array([event.delay])
            delay_id, found = self.delay_library.find(data)
            if not found:
                self.delay_library.insert(delay_id, data)
            self.block_events[block_index][0] = delay_id
            duration = max(duration, event.delay)

    # =========
    # PERFORM GRADIENT CHECKS
    # =========
    for cg_temp in check_g.keys():
        cg = check_g[cg_temp]

        if abs(cg.start[1]) > self.system.max_slew * self.system.grad_raster_time:
            if cg.start[0] != 0:
                raise ValueError('No delay allowed for gradients which start with a non-zero amplitude')

            if block_index > 1:
                prev_id = self.block_events[block_index - 1][cg.idx]
                if prev_id != 0:
                    prev_lib = self.grad_library.get(prev_id)
                    prev_dat = prev_lib['data']
                    prev_type = prev_lib['type']
                    if prev_type == 't':
                        raise Exception(
                            'Two consecutive gradients need to have the same amplitude at the connection point')
                    elif prev_type == 'g':
                        last = prev_dat[4]
                        if abs(last - cg.start[1]) > self.system.max_slew * self.system.grad_raster_time:
                            raise Exception(
                                'Two consecutive gradients need to have the same amplitude at the connection point')
            else:
                raise Exception('First gradient in the the first block has to start at 0.')
        if cg.stop[1] > self.system.max_slew * self.system.grad_raster_time and abs(cg.stop[0] - block_duration) > 1e-7:
            raise Exception('A gradient that doesnt end at zero needs to be aligned to the block boundary.')


def get_block(self, block_index: int) -> SimpleNamespace:
    """
    Returns Pulseq block at `block_index` position in `self.block_events`.

    Parameters
    ----------
    block_index : int
        Index of Pulseq block to be retrieved from `self.block_events`.

    Returns
    -------
    block : SimpleNamespace
        Pulseq block at 'block_index' position in `self.block_events`.
    """

    block = SimpleNamespace()
    event_ind = self.block_events[block_index]
    if event_ind[0] > 0:
        delay = SimpleNamespace()
        delay.type = 'delay'
        delay.delay = self.delay_library.data[event_ind[0]][0]
        block.delay = delay
    elif event_ind[1] > 0:
        block.rf = self.rf_from_lib_data(self.rf_library.data[event_ind[1]])
    grad_channels = ['gx', 'gy', 'gz']
    for i in range(1, len(grad_channels) + 1):
        if event_ind[2 + (i - 1)] > 0:
            grad, compressed = SimpleNamespace(), SimpleNamespace()
            type = self.grad_library.type[event_ind[2 + (i - 1)]]
            lib_data = self.grad_library.data[event_ind[2 + (i - 1)]]
            grad.type = 'trap' if type == 't' else 'grad'
            grad.channel = grad_channels[i - 1][1]
            if grad.type == 'grad':
                amplitude = lib_data[0]
                shape_id = lib_data[1]
                delay = lib_data[2]
                shape_data = self.shape_library.data[shape_id]
                compressed.num_samples = shape_data[0]
                compressed.data = shape_data[1:]
                g = decompress_shape(compressed)
                grad.waveform = amplitude * g
                grad.t = np.arange(g.size) * self.grad_raster_time
                grad.delay = delay
                if len(lib_data) > 4:
                    grad.first = lib_data[3]
                    grad.last = lib_data[4]
                else:
                    grad.first = grad.waveform[0]
                    grad.last = grad.waveform[-1]
            else:
                if max(lib_data.shape) < 5: # added by GT
                    grad.amplitude, grad.rise_time, grad.flat_time, grad.fall_time = [lib_data[x] for x in range(4)]
                    grad.delay = 0
                else:
                    grad.amplitude, grad.rise_time, grad.flat_time, grad.fall_time, grad.delay = [lib_data[x] for x in
                                                                                              range(5)]
                grad.area = grad.amplitude * (grad.flat_time + grad.rise_time / 2 + grad.fall_time / 2)
                grad.flat_area = grad.amplitude * grad.flat_time
            setattr(block, grad_channels[i - 1], grad)

    if event_ind[5] > 0:
        lib_data = self.adc_library.data[event_ind[5]]
        if max(lib_data.shape) < 6:
            lib_data = np.append(lib_data, 0)
        adc = SimpleNamespace()
        adc.num_samples, adc.dwell, adc.delay, adc.freq_offset, adc.phase_offset, adc.dead_time = [lib_data[x] for x in
                                                                                                   range(6)]
        adc.num_samples = int(adc.num_samples)
        adc.type = 'adc'
        block.adc = adc
    return block



from math import pi


def convert(from_value: float, from_unit: str, to_unit: str = None, gamma: float = 42.576e6) -> float:
    """"
    Converts gradient amplitude or slew rate from unit `from_unit` to unit `to_unit` with gyromagnetic ratio `gamma`.

    Parameters
    ----------
    from_value : float
        Gradient amplitude or slew rate to convert from.
    from_unit : str
        Unit of gradient amplitude or slew rate to convert from.
    to_unit : str, optional
        Unit of gradient amplitude or slew rate to convert to. Default is `None`.
    gamma : float, optional
        Gyromagnetic ratio. Default is 42.576e6, for Hydrogen.

    Returns
    -------
    out : float
        Converted gradient amplitude or slew rate.
    """
    valid_grad_units = ['Hz/m', 'mT/m', 'rad/ms/mm']
    valid_slew_units = ['Hz/m/s', 'mT/m/ms', 'T/m/s', 'rad/ms/mm/ms']
    valid_units = valid_grad_units
    valid_units.extend(valid_slew_units)

    if from_unit not in valid_grad_units:
        raise ValueError()

    if to_unit is not None and to_unit not in valid_units:
        raise ValueError()

    if to_unit is None:
        if from_unit in valid_grad_units:
            to_unit = valid_grad_units[0]
        elif from_unit in valid_slew_units:
            to_unit = valid_slew_units[0]

    # Convert to standard units
    # Grad units
    if from_unit == 'Hz/m':
        standard = from_value
    elif from_unit == 'mT/m':
        standard = from_value * 1e-3 * gamma
    elif from_unit == 'rad/ms/mm':
        standard = from_value * 1e6 / (2 * pi)
    # Slew units
    elif from_unit == 'Hz/m/s':
        standard = from_value
    elif from_unit == 'mT/m/ms' or from_unit == 'T/m/s':
        standard = from_value * gamma
    elif from_unit == 'rad/ms/mm/ms':
        standard = from_value * 1e9 / (2 * pi)

    # Convert from standard units
    # Grad units
    if to_unit == 'Hz/m':
        out = standard
    elif to_unit == 'mT/m':
        out = 1e3 * standard / gamma
    elif to_unit == 'rad/ms/mm':
        out = standard * 2 * pi * 1e-6
    # Slew units
    elif to_unit == 'Hz/m/s':
        out = standard
    elif to_unit == 'mT/m/ms' or to_unit == 'T/m/s':
        out = standard / gamma
    elif to_unit == 'rad/ms/mm/ms':
        out = standard * 2 * pi * 1e-9

    return out



import numpy as np


def write(self, file_name):
    """
    Writes the calling `Sequence` object as a `.seq` file with filename `file_name`.

    Parameters
    ----------
    file_name : str
        File name of `.seq` file to be written to disk.
    """
    # `>.0f` is used when only decimals have to be displayed.
    # `>g` is used when insignificant zeros have to be truncated.
    file_name += '.seq' if file_name[-4:] != '.seq' not in file_name else ''
    output_file = open(file_name, 'w')
    output_file.write('# Pulseq sequence file\n')
    output_file.write('# Created by PyPulseq\n\n')

    output_file.write('[VERSION]\n')
    output_file.write(f'major {self.version_major}\n')
    output_file.write(f'minor {self.version_minor}\n')
    output_file.write(f'revision {self.version_revision}\n')
    output_file.write('\n')

    if len(self.definitions) != 0:
        output_file.write('[DEFINITIONS]\n')
        keys = list(self.definitions.keys())
        values = list(self.definitions.values())
        for i in range(len(keys)):
            output_file.write(f'{keys[i]} ')
            if isinstance(values[i], str):
                output_file.write(f'{values[i]} ')
            else:
                for t in range(len(values[i])):
                    output_file.write(f'{values[i][t]:.6g} ')
            output_file.write('\n')
        output_file.write('\n')

    output_file.write('# Format of blocks:\n')
    output_file.write('#  #  D RF  GX  GY  GZ ADC\n')
    output_file.write('[BLOCKS]\n')
    id_format_width = '{:' + str(len(str(len(self.block_events)))) + 'd} '
    id_format_str = id_format_width + '{:2d} {:2d} {:3d} {:3d} {:3d} {:2d}\n'
    for i in range(len(self.block_events)):
        s = id_format_str.format(*np.insert(self.block_events[i + 1], 0, (i + 1)))
        output_file.write(s)
    output_file.write('\n')

    if len(self.rf_library.keys) != 0:
        output_file.write('# Format of RF events:\n')
        output_file.write('# id amplitude mag_id phase_id delay freq phase\n')
        output_file.write('# ..        Hz   ....     ....    us   Hz   rad\n')
        output_file.write('[RF]\n')
        rf_lib_keys = self.rf_library.keys
        # id_format_str = '{:>1.0f} {:>12g} {:>1.0f} {:>1.0f} {:>g} {:>g}\n'
        id_format_str = '{:.0f} {:12g} {:.0f} {:.0f} {:g} {:g} {:g}\n'
        for k in rf_lib_keys.keys():
            lib_data1 = self.rf_library.data[k][0:3]
            lib_data2 = self.rf_library.data[k][4:7]
            delay = round(self.rf_library.data[k][3] * 1e6)
            s = id_format_str.format(k, *lib_data1, delay, *lib_data2)
            output_file.write(s)
        output_file.write('\n')

    grad_lib_values = np.array(list(self.grad_library.type.values()))
    arb_grad_mask = np.where(grad_lib_values == 'g')[0]
    trap_grad_mask = np.where(grad_lib_values == 't')[0]

    if len(arb_grad_mask) != 0:
        output_file.write('# Format of arbitrary gradients:\n')
        output_file.write('# id amplitude shape_id delay\n')
        output_file.write('# ..      Hz/m     ....    us\n')
        output_file.write('[GRADIENTS]\n')
        id_format_str = '{:.0f} {:12g} {:.0f} {:.0f}\n'
        keys = np.array(list(self.grad_library.keys.keys()))
        for k in keys[arb_grad_mask]:
            s = id_format_str.format(k, *self.grad_library.data[k][:2], round(self.grad_library.data[k][2] * 1e6))
            output_file.write(s)
        output_file.write('\n')

    if len(trap_grad_mask) != 0:
        output_file.write('# Format of trapezoid gradients:\n')
        output_file.write('# id amplitude rise flat fall delay\n')
        output_file.write('# ..      Hz/m   us   us   us    us\n')
        output_file.write('[TRAP]\n')
        keys = np.array(list(self.grad_library.keys.keys()))
        id_format_str = '{:2g} {:12g} {:3g} {:4g} {:3g} {:3g}\n'
        for k in keys[trap_grad_mask]:
            data = np.copy(self.grad_library.data[k])  # Make a copy to leave the original untouched
            data[1:] = np.round(1e6 * data[1:])
            s = id_format_str.format(k, *data)
            output_file.write(s)
        output_file.write('\n')

    if len(self.adc_library.keys) != 0:
        output_file.write('# Format of ADC events:\n')
        output_file.write('# id num dwell delay freq phase\n')
        output_file.write('# ..  ..    ns    us   Hz   rad\n')
        output_file.write('[ADC]\n')
        keys = self.adc_library.keys
        id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f} {:.0f} {:.6g}\n'
        for k in keys.values():
            data = np.multiply(self.adc_library.data[k][0:5], [1, 1e9, 1e6, 1, 1])
            s = id_format_str.format(k, *data)
            output_file.write(s)
        output_file.write('\n')

    if len(self.delay_library.keys) != 0:
        output_file.write('# Format of delays:\n')
        output_file.write('# id delay (us)\n')
        output_file.write('[DELAYS]\n')
        keys = self.delay_library.keys
        id_format_str = '{:.0f} {:.0f}\n'
        for k in keys.values():
            s = id_format_str.format(k, *np.round(1e6 * self.delay_library.data[k]))
            output_file.write(s)
        output_file.write('\n')

    if len(self.shape_library.keys) != 0:
        output_file.write('# Sequence Shapes\n')
        output_file.write('[SHAPES]\n\n')
        keys = self.shape_library.keys
        for k in keys.values():
            shape_data = self.shape_library.data[k]
            s = 'shape_id {:.0f}\n'
            s = s.format(k)
            output_file.write(s)
            s = 'num_samples {:.0f}\n'
            s = s.format(shape_data[0])
            output_file.write(s)
            s = '{:.9g}\n' * len(shape_data[1:])
            s = s.format(*shape_data[1:])
            output_file.write(s)
            output_file.write('\n')



from types import SimpleNamespace


def calc_duration(*events: list) -> float:
    """
    Calculate the cumulative duration of Events.

    Parameters
    ----------
    events : list
        List of `SimpleNamespace` events. Can also be a list containing a single block (see
        `pypulseq.Sequence.sequence.plot()`).

    Returns
    -------
    duration : float
        The cumulative duration of the pulse events in `events`.
    """

    for e in events:
        if isinstance(e, SimpleNamespace) and isinstance(getattr(e, list(e.__dict__.keys())[0]), SimpleNamespace):
            events = list(e.__dict__.values())
            break

    duration = 0
    for event in events:
        if not isinstance(event, SimpleNamespace):
            raise TypeError("input(s) should be of type SimpleNamespace")

        if event.type == 'delay':
            duration = max(duration, event.delay)
        elif event.type == 'rf':
            duration = max(duration, event.delay + event.t[-1])
        elif event.type == 'grad':
            duration = max(duration, event.t[-1] + event.t[1] - event.t[0] + event.delay)
        elif event.type == 'adc':
            duration = max(duration, event.delay + event.num_samples * event.dwell + event.dead_time)
        elif event.type == 'trap':
            duration = max(duration, event.delay + event.rise_time + event.flat_time + event.fall_time)

    return round(duration, ndigits=6)



from types import SimpleNamespace

import numpy as np


def calc_rf_center(rf: SimpleNamespace):
    """
    Calculate the time point of the effective rotation calculated as the peak of the radio-frequency amplitude for the
    shaped pulses and the center of the pulse for the block pulses. Zero padding in the radio-frequency pulse is
    considered as a part of the shape. Delay field of the radio-frequency object is not taken into account.

    Parameters
    ----------
    rf : SimpleNamespace
        Radio-frequency pulse event.

    Returns
    -------
    tc : float
        Time point of the center of the radio-frequency pulse.
    ic : float
        Corresponding position of `tc` in the radio-frequency pulse's envelope.
    """
    eps = np.finfo(float).eps
    for first, x in enumerate(rf.signal):
        if abs(x) > eps:
            break

    for last, x in enumerate(rf.signal[::-1]):
        if abs(x) > eps:
            break

    last = len(rf.signal) - last - 1  # Because we traverse over in-place reverse of rf.signal
    rf_min = min(abs(rf.signal[first:last]))
    rf_max, ic = max(abs(rf.signal[first:last])), np.argmax(abs(rf.signal[first:last]))
    if rf_max - rf_min <= eps:
        ic = round((last - first + 1) / 2)

    tc = rf.t[first - 1 + ic]

    return tc, ic



from pypulseq.convert import convert


class Opts:
    """
    System limits of an MR scanner.

    Attributes
    ----------
    grad_unit : str
        Unit of maximum gradient amplitude. Must be one of Hz/m, mT/m or rad/ms/mm.
    slew_unit : str
        Unit of maximum slew rate. Must be one of Hz/m/s, mT/m/ms, T/m/s or rad/ms/mm/ms.
    max_grad : float
        Maximum gradient amplitude.
    max_slew : float
        Maximum slew rate.
    rise_time : float
        Rise time for gradients.
    rf_dead_time : float
        Dead time for radio-frequency pulses.
    rf_ringdown_time : float
        Ringdown time for radio-frequency pulses.
    adc_dead_time : float
        Dead time for ADC readout pulses.
    rf_raster_time : float
        Raster time for radio-frequency pulses.
    grad_raster_time : float
        Raster time for gradient waveforms.
    gamma : float
        Gyromagnetic ratio. Default is 42.576 MHz for Hydrogen.
    """
    def __init__(self, grad_unit: str = None, slew_unit: str = None, max_grad: float = None, max_slew: float = None,
                 rise_time: float = None, rf_dead_time: float = 0, rf_ringdown_time: float = 0,
                 adc_dead_time: float = 0, rf_raster_time: float = 1e-6, grad_raster_time: float = 10e-6,
                 gamma: float = 42.576e6):
        valid_grad_units = ['Hz/m', 'mT/m', 'rad/ms/mm']
        valid_slew_units = ['Hz/m/s', 'mT/m/ms', 'T/m/s', 'rad/ms/mm/ms']

        if grad_unit is not None and grad_unit not in valid_grad_units:
            raise ValueError(f'Invalid gradient unit, must be one of Hz/m, mT/m or rad/ms/mm. You passed: {grad_unit}')

        if slew_unit is not None and slew_unit not in valid_slew_units:
            raise ValueError(f'Invalid slew rate unit, must be one of Hz/m/s, mT/m/ms, T/m/s or rad/ms/mm/ms. You '
                             f'passed: {slew_unit}')

        if max_grad is None:
            max_grad = convert(from_value=40, from_unit='mT/m', gamma=gamma)
        else:
            max_grad = convert(from_value=max_grad, from_unit=grad_unit, to_unit='Hz/m', gamma=gamma)

        if max_slew is None:
            max_slew = convert(from_value=170, from_unit='T/m/s', gamma=gamma)
        else:
            max_slew = convert(from_value=max_slew, from_unit=slew_unit, to_unit='Hz/m', gamma=gamma)

        if rise_time is not None:
            max_slew = None

        self.max_grad = max_grad
        self.max_slew = max_slew
        self.rise_time = rise_time
        self.rf_dead_time = rf_dead_time
        self.rf_ringdown_time = rf_ringdown_time
        self.adc_dead_time = adc_dead_time
        self.rf_raster_time = rf_raster_time
        self.grad_raster_time = grad_raster_time
        self.gamma = gamma

    def __str__(self):
        s = "System limits:"
        s += "\nmax_grad: " + str(self.max_grad) + str(self.grad_unit)
        s += "\nmax_slew: " + str(self.max_slew) + str(self.slew_unit)
        s += "\nrise_time: " + str(self.rise_time)
        s += "\nrf_dead_time: " + str(self.rf_dead_time)
        s += "\nrf_ring_time: " + str(self.rf_ringdown_time)
        s += "\nadc_dead_time: " + str(self.adc_dead_time)
        s += "\nrf_raster_time: " + str(self.rf_raster_time)
        s += "\ngrad_raster_time: " + str(self.grad_raster_time)
        s += "\ngamma: " + str(self.gamma)
        return s



from types import SimpleNamespace


def make_delay(d: float) -> SimpleNamespace:
    """
    Creates a delay event.

    Parameters
    ----------
    d : float
        Delay time in milliseconds (ms).

    Returns
    -------
    delay : SimpleNamespace
        Delay event.
    """

    delay = SimpleNamespace()
    d = round(d, 5)  # causes crashes on scanners if the precision is better than 10us - this is confusing - To Do:
    if d < 0:
        raise ValueError('Delay {:.2f} ms is invalid'.format(d * 1e3))
    delay.type = 'delay'
    delay.delay = d
    return delay



import math
from types import SimpleNamespace

from pypulseq.opts import Opts


def make_trapezoid(channel='z', system=Opts(), duration=0, area=-1, flat_time=None, flat_area=None, amplitude=None,
                   max_grad=0, max_slew=0, rise_time=0, delay=0):
    """
    Creates a trapezoidal gradient event.

    Parameters
    ----------
    channel : str, optional
        Orientation of trapezoidal gradient event. Must be one of `x`, `y` or `z`.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    duration : float, optional
        Duration in milliseconds (ms). Default is 0.
    area : float, optional
        Area. Default is -1.
    flat_time : float, optional
        Flat duration in milliseconds (ms). Default is `None`.
    flat_area : float, optional
        Flat area. Default is `None`.
    amplitude : float, optional
        Amplitude. Default is `None`.
    max_grad : float, optional
        Maximum gradient strength. Default is 0.
    max_slew : float, optional
        Maximum slew rate. Default is 0.
    rise_time : float, optional
        Rise time in milliseconds (ms). Default is 0.
    delay : float, optional
        Delay in milliseconds (ms). Default is 0.

    Returns
    -------
    grad : SimpleNamespace
        Trapezoidal gradient event created based on the supplied parameters.
    """
    if channel not in ['x', 'y', 'z']:
        raise ValueError()

    if max_grad <= 0:
        max_grad = system.max_grad

    if max_slew <= 0:
        max_slew = system.max_slew

    if rise_time <= 0:
        rise_time = system.rise_time

    if area is None and flat_area is None and amplitude is None:
        raise ValueError('Must supply either \'area\', \'flat_area\' or \'amplitude\'')

    if flat_time is not None:
        if amplitude is not None:
            amplitude2 = amplitude
        else:
            amplitude2 = flat_area / flat_time

        if rise_time is None:
            rise_time = abs(amplitude2) / max_slew
            rise_time = math.ceil(rise_time / system.grad_raster_time) * system.grad_raster_time
        fall_time, flat_time = rise_time, flat_time
    elif duration > 0:
        amplitude2 = amplitude
        if amplitude is None:
            if rise_time is None:
                dC = 1 / abs(2 * max_slew) + 1 / abs(2 * max_slew)
                possible = duration ** 2 > 4 * abs(area) * dC
                amplitude2 = (duration - math.sqrt(duration ** 2 - 4 * abs(area) * dC)) / (2 * dC)
            else:
                amplitude2 = area / (duration - rise_time)
                possible = duration > 2 * rise_time and abs(amplitude2) < max_grad

            if not possible:
                raise ValueError('Requested area is too large for this gradient')

        if rise_time is None:
            rise_time = math.ceil(
                abs(amplitude2) / max_slew / system.grad_raster_time) * system.grad_raster_time

        fall_time = rise_time
        flat_time = duration - rise_time - fall_time

        if amplitude is None:
            amplitude2 = area / (rise_time / 2 + fall_time / 2 + flat_time)
    else:
        if area is None:
            raise ValueError('Must supply a duration')
        else:
            rise_time = math.ceil(math.sqrt(abs(area) / max_slew) / system.grad_raster_time) * system.grad_raster_time
            amplitude2 = area / rise_time
            t_eff = rise_time

            if abs(amplitude2) > max_grad:
                t_eff = math.ceil(abs(area) / max_grad / system.grad_raster_time) * system.grad_raster_time
                amplitude2 = area / t_eff
                rise_time = math.ceil(
                    abs(amplitude2) / max_slew / system.grad_raster_time) * system.grad_raster_time

            flat_time = t_eff - rise_time
            fall_time = rise_time

    if abs(amplitude2) > max_grad:
        raise ValueError("Amplitude violation")

    grad = SimpleNamespace()
    grad.type = 'trap'
    grad.channel = channel
    grad.amplitude = amplitude2
    grad.rise_time = rise_time
    grad.flat_time = flat_time
    grad.fall_time = fall_time
    grad.area = amplitude2 * (flat_time + rise_time / 2 + fall_time / 2)
    grad.flat_area = amplitude2 * flat_time
    grad.delay = delay
    grad.first = 0
    grad.last = 0

    return grad



import numpy as np
from pypulseq.convert import convert


def test_report(self):
    """
    Analyze the sequence and return a text report.
    """
    # Find RF pulses and list flip angles
    flip_angles_deg = []
    for k in self.rf_library.keys:
        lib_data = self.rf_library.data[k]
        rf = self.rf_from_lib_data(lib_data)
        flip_angles_deg.append(np.abs(np.sum(rf.signal) * rf.t[0] * 360))

    flip_angles_deg = np.unique(flip_angles_deg)

    # Calculate TE, TR
    duration, num_blocks, event_count = self.duration()

    k_traj_adc, k_traj, t_excitation, t_refocusing, t_adc = self.calculate_kspace()

    k_abs_adc = np.sqrt(np.sum(np.square(k_traj_adc), axis=0))
    k_abs_echo, index_echo = np.min(k_abs_adc), np.argmin(k_abs_adc)
    t_echo = t_adc[index_echo]
    t_ex_tmp = t_excitation[t_excitation < t_echo]
    TE = t_echo - t_ex_tmp[-1]

    if len(t_excitation) < 2:
        TR = duration
    else:
        t_ex_tmp1 = t_excitation[t_excitation > t_echo]
        if len(t_ex_tmp1) == 0:
            TR = t_ex_tmp[-1] - t_ex_tmp[-2]
        else:
            TR = t_ex_tmp1[0] - t_ex_tmp[-1]

    # Check sequence dimensionality and spatial resolution
    k_extent = np.max(np.abs(k_traj_adc), axis=1)
    k_scale = np.max(k_extent)
    if k_scale != 0:
        k_bins = 4e6
        k_threshold = k_scale / k_bins

        # Detect unused dimensions and delete them
        if np.any(k_extent < k_threshold):
            k_traj_adc = np.delete(k_traj_adc, np.where(k_extent < k_threshold), axis=0)
            k_extent = np.delete(k_extent, np.where(k_extent < k_threshold), axis=0)

        # Bin the k-space trajectory to detect repetitions / slices
        k_len = k_traj_adc.shape[1]
        k_repeat = np.zeros(k_len)

        k_map = dict()
        for i in range(k_len):
            l = k_bins + np.round(k_traj_adc[:, i] / k_threshold)
            key_string = ('{:.0f} ' * len(l)).format(*l)
            if key_string in k_map:
                k_repeat[i] = k_map[key_string] + 1
            else:
                k_repeat[i] = 1
            k_map[key_string] = k_repeat[i]

        repeats = np.max(k_repeat)

        k_traj_rep1 = k_traj_adc[:, k_repeat == 1]

        k_counters = np.zeros(k_traj_rep1.shape)
        dims = k_traj_rep1.shape[0]
        ordering = dict()
        for j in range(dims):
            c = 1
            k_map = dict()
            for i in range(k_traj_rep1.shape[1]):
                key = round(k_traj_rep1[j, i] / k_threshold)
                if key in k_map:
                    k_counters[j, i] = k_map[key]
                else:
                    k_counters[j, i] = c
                    k_map[key] = c
                    c += 1
            ordering[j] = k_map.values()

        unique_k_positions = np.max(k_counters, axis=1)
        is_cartesian = np.prod(unique_k_positions) == k_traj_rep1.shape[1]
    else:
        unique_k_positions = 1

    gw = self.gradient_waveforms()
    gws = (gw[:, 1:] - gw[:, :-1]) / self.system.grad_raster_time
    ga = np.max(np.abs(gw), axis=1)
    gs = np.max(np.abs(gws), axis=1)

    ga_abs = np.max(np.sqrt(np.sum(np.square(gw), axis=0)))
    gs_abs = np.max(np.sqrt(np.sum(np.square(gws), axis=0)))

    timing_ok, timing_error_report = self.check_timing()

    report = f'Number of blocks: {num_blocks}\n' \
        f'Number of events:\n' \
        f'RF: {event_count[1]:6.0f}\n' \
        f'Gx: {event_count[2]:6.0f}\n' \
        f'Gy: {event_count[3]:6.0f}\n' \
        f'Gz: {event_count[4]:6.0f}\n' \
        f'ADC: {event_count[5]:6.0f}\n' \
        f'Delay: {event_count[0]:6.0f}\n' \
        f'Sequence duration: {duration:.6f} s\n' \
        f'TE: {TE:.6f} s\n' \
        f'TR: {TR:.6f} s\n'
    report += 'Flip angle: ' + ('{:.02f} ' * len(flip_angles_deg)).format(*flip_angles_deg) + 'deg\n'
    report += 'Unique k-space positions (aka cols, rows, etc.): ' + ('{:.0f} ' * len(unique_k_positions)).format(
        *unique_k_positions) + '\n'

    if len(np.where(unique_k_positions > 1)):
        report += f'Dimensions: {len(k_extent)}\n'
        report += ('Spatial resolution: {:.02f} mm\n' * len(k_extent)).format(*(0.5 / k_extent * 1e3))
        report += f'Repetitions/slices/contrasts: {repeats}\n'

        if is_cartesian:
            report += 'Cartesian encoding trajectory detected\n'
        else:
            report += 'Non-cartesian/irregular encoding trajectory detected (eg: EPI, spiral, radial, etc.)\n'

    if timing_ok:
        report += 'Block timing check passed successfully\n'
    else:
        report += f'Block timing check failed. Error listing follows:\n {timing_error_report}'

    ga_converted = convert(from_value=ga, from_unit='Hz/m', to_unit='mT/m')
    gs_converted = convert(from_value=gs, from_unit='Hz/m/s', to_unit='T/m/s')
    report += 'Max gradient: ' + ('{:.0f} ' * len(ga)).format(*ga) + 'Hz/m == ' + (
            '{:.02f} ' * len(ga_converted)).format(*ga_converted) + 'mT/m\n'
    report += 'Max slew rate: ' + ('{:.0f} ' * len(gs)).format(*gs) + 'Hz/m/s == ' + (
            '{:.02f} ' * len(gs_converted)).format(*gs_converted) + 'mT/m/s\n'

    ga_abs_converted = convert(from_value=ga_abs, from_unit='Hz/m', to_unit='mT/m')
    gs_abs_converted = convert(from_value=gs_abs, from_unit='Hz/m/s', to_unit='T/m/s')
    report += f'Max absolute gradient: {ga_abs:.0f} Hz/m == {ga_abs_converted:.2f} mT/m\n'
    report += f'Max absolute slew rate: {gs_abs:g} Hz/m/s == {gs_abs_converted:.2f} T/m/s'

    return report



import numpy as np


class EventLibrary:
    """
    Defines an event library. Provides methods to insert new data and find existing data.

    Attributes
    ----------
    keys : dict
        Key-value pairs of event keys and corresponding... event keys.
    data : dict
        Key-value pairs of event keys and corresponding data.
    lengths : dict
        Key-value pairs of event keys and corresponding length of data values in `self.data`.
    type : dict
        Key-value pairs of event keys and corresponding event types.
    keymap : dict
        Key-value pairs of data values and corresponding event keys.
    """

    def __init__(self):
        self.keys, self.data, self.lengths, self.type, self.keymap = dict(), dict(), dict(), dict(), dict()

    def __str__(self):
        s = "EventLibrary:"
        s += "\nkeys: " + str(len(self.keys))
        s += "\ndata: " + str(len(self.data))
        s += "\nlengths: " + str(len(self.lengths))
        s += "\ntype: " + str(len(self.type))
        return s

    def find(self, new_data: np.ndarray):
        """
        Finds data `new_data` in event library.

        Parameters
        ----------
        new_data : np.ndarray
            Data to be found in event library.

        Returns
        -------
        key_id : int
            Key of `new_data` in event library, if found.
        found : bool
            If `new_data` was found in the event library or not.
        """
        data_string = np.array2string(new_data, formatter={'float': lambda x: f'{x:.6g}'})
        data_string = data_string.replace('[', '')
        data_string = data_string.replace(']', '')
        try:
            key_id = self.keymap[data_string]
            found = True
        except:
            key_id = 1 if len(self.keys) == 0 else max(self.keys) + 1
            found = False

        return key_id, found

    def insert(self, key_id: int, new_data: np.ndarray, data_type: str = None):
        """
        Inserts `new_data` of data type `data_type` into the event library with key `key_id`.

        Parameters
        ----------
        key_id : int
            Key of `new_data`.
        new_data : np.ndarray
            Data to be inserted into event library.
        data_type : str
            Data type of `new_data`.
        """
        self.keys[key_id] = key_id
        self.data[key_id] = new_data
        self.lengths[key_id] = max(new_data.shape)
        data_string = np.array2string(new_data, formatter={'float_kind': lambda x: "%.6g" % x})
        data_string = data_string.replace('[', '')
        data_string = data_string.replace(']', '')
        self.keymap[data_string] = key_id
        if data_type is not None:
            self.type[key_id] = data_type

    def get(self, key_id: int):
        return {'key': self.keys[key_id], 'data': self.data[key_id], 'length': self.lengths[key_id],
                'type': self.type[key_id]}



import numpy as np


def decompress_shape(compressed_shape: np.ndarray) -> np.ndarray:
    """
    Decompresses a run-length encoded shape.

    Parameters
    ----------
    compressed_shape : numpy.ndarray
        Run-length encoded shape.

    Returns
    -------
    decompressed_shape : numpy.ndarray
        Decompressed shape.
    """
    data_pack, num_samples = compressed_shape.data, int(compressed_shape.num_samples)
    decompressed_shape = np.zeros(num_samples)

    count_pack, count_unpack = 0, 0
    while count_pack < max(data_pack.shape) - 1:
        if data_pack[count_pack] != data_pack[count_pack + 1]:
            decompressed_shape[count_unpack] = data_pack[count_pack]
            count_unpack += 1
            count_pack += 1
        else:
            rep = int(data_pack[count_pack + 2] + 2)
            decompressed_shape[count_unpack:(count_unpack + rep)] = data_pack[count_pack]
            count_pack += 3
            count_unpack += rep

    if count_pack == max(data_pack.shape) - 1:
        decompressed_shape[count_unpack] = data_pack[count_pack]

    decompressed_shape = np.cumsum(decompressed_shape)
    return decompressed_shape



from types import SimpleNamespace

import numpy as np


def compress_shape(decompressed_shape: np.ndarray) -> np.ndarray:
    """
    Returns a run-length encoded compressed shape.

    Parameters
    ----------
    decompressed_shape : numpy.ndarray
        Decompressed shape.

    Returns
    -------
    compressed_shape : SimpleNamespace
        A `SimpleNamespace` object containing the compressed data and corresponding shape.
    """
    quant_factor = 1e-8
    decompressed_shape_scaled = decompressed_shape / quant_factor
    datq = np.round(np.insert(np.diff(decompressed_shape_scaled), 0, decompressed_shape_scaled[0]))
    qerr = decompressed_shape_scaled - np.cumsum(datq)
    qcor = np.insert(np.diff(np.round(qerr)), 0, 0)
    datd = datq + qcor
    mask_changes = np.insert(np.asarray(np.diff(datd) != 0, dtype=np.int), 0, 1)
    vals = datd[mask_changes.nonzero()[0]] * quant_factor

    k = np.append(mask_changes, 1).nonzero()[0]
    n = np.diff(k)

    n_extra = (n - 2).astype(np.float32)  # Cast as float for nan assignment to work
    vals2 = np.copy(vals)
    vals2[n_extra < 0] = np.nan
    n_extra[n_extra < 0] = np.nan
    v = np.stack((vals, vals2, n_extra))
    v = v.T[np.isfinite(v).T]  # Use transposes to match Matlab's Fortran indexing order
    v[abs(v) < 1e-10] = 0
    compressed_shape = SimpleNamespace()
    compressed_shape.num_samples = len(decompressed_shape)
    compressed_shape.data = v

    return compressed_shape



import numpy as np

from pypulseq.opts import Opts


def traj_to_grad(k: np.ndarray, raster_time: float = Opts().grad_raster_time):
    """
    Convert k-space trajectory `k` into gradient waveform in compliance with `raster_time` gradient raster time.

    Parameters
    ----------
    k : numpy.ndarray
        K-space trajectory to be converted into gradient waveform.
    raster_time : float, optional
        Gradient raster time. Default is the default value of gradient raster time set in system limits `Opts`.

    Returns
    -------
    g : numpy.ndarray
        Gradient waveform.
    sr : numpy.ndarray
        Slew rate.
    """
    g = (k[1:] - k[:-1]) / raster_time
    sr0 = (g[1:] - g[:-1]) / raster_time
    sr = np.zeros(len(sr0) + 1)
    sr[0] = sr0[0]
    sr[1:-1] = 0.5 * (sr0[-1] + sr0[1:])
    sr[-1] = sr0[-1]

    return g, sr



from types import SimpleNamespace

import numpy as np

from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.opts import Opts


def split_gradient_at(grad: SimpleNamespace, time_point: float, system: Opts = Opts()):
    """
    Split gradient waveform `grad` into two at time point `time_point`.

    Parameters
    ----------
    grad : SimpleNamespace
        Gradient event to be split into two gradient events.
    time_point : float, optional
        Time point at which `grad` will be split into two gradient waveforms.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    Returns
    -------
    grad1, grad2 : numpy.ndarray
        Gradient waveforms after splitting.
    """
    grad_raster_time = system.grad_raster_time

    time_index = round(time_point / grad_raster_time)
    time_point = round(time_index * grad_raster_time, 6)  # Work around floating-point arithmetic limitation
    time_index += 1

    if grad.type == 'trap':
        ch = grad.channel
        grad.delay = round(grad.delay / grad_raster_time) * grad_raster_time
        grad.rise_time = round(grad.rise_time / grad_raster_time) * grad_raster_time
        grad.flat_time = round(grad.flat_time / grad_raster_time) * grad_raster_time
        grad.fall_time = round(grad.fall_time / grad_raster_time) * grad_raster_time

        if grad.flat_time == 0:
            times = [0, grad.rise_time, grad.rise_time + grad.fall_time]
            amplitudes = [0, grad.amplitude, 0]
        else:
            times = [0, grad.rise_time, grad.rise_time + grad.flat_time,
                     grad.rise_time + grad.flat_time + grad.fall_time]
            amplitudes = [0, grad.amplitude, grad.amplitude, 0]

        if time_point < grad.delay:
            times = np.insert(grad.delay + times, 0, 0)
            amplitudes = [0, amplitudes]
            grad.delay = 0

        amplitudes = np.array(amplitudes)
        times = np.array(times).round(6)  # Work around floating-point arithmetic limitation

        amp_tp = np.interp(x=time_point, xp=times, fp=amplitudes)
        times1 = np.append(times[np.where(times < time_point)], time_point)
        amplitudes1 = np.append(amplitudes[np.where(times < time_point)], amp_tp)
        times2 = np.insert(times[times > time_point], 0, time_point) - time_point
        amplitudes2 = np.insert(amplitudes[times > time_point], 0, amp_tp)

        grad1 = make_extended_trapezoid(channel=ch, system=system, times=times1, amplitudes=amplitudes1,
                                        skip_check=True)
        grad1.delay = grad.delay
        grad2 = make_extended_trapezoid(channel=ch, system=system, times=times2, amplitudes=amplitudes2,
                                        skip_check=True)
        grad2.delay = time_point
        return grad1, grad2
    elif grad.type == 'grad':
        if time_index == 1 or time_index >= len(grad.t):
            return grad
        else:
            grad1 = grad
            grad2 = grad
            grad1.last = grad.waveform[time_index]
            grad2.first = grad.waveform[time_index]
            grad2.delay = grad.delay + grad.t[time_index]
            grad1.t = grad.t[:time_index]
            grad1.waveform = grad.waveform[:time_index]
            grad2.t = grad.t[time_index:]
            grad2.waveform = grad.waveform[time_index:]
            return grad1, grad2
    else:
        raise ValueError('Splitting of unsupported event.')



from types import SimpleNamespace

import numpy as np

from pypulseq.make_arbitrary_grad import make_arbitrary_grad
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform


def make_extended_trapezoid(channel: str, times: np.ndarray = np.zeros(1), amplitudes: np.ndarray = np.zeros(1),
                            system: Opts = Opts(), max_grad: float = 0, max_slew: float = 0, skip_check: bool = False):
    """
    Creates an extend trapezoidal gradient event by defined by amplitude values in `amplitudes` at time indices in
    `times`.

    Parameters
    ----------
    channel : str
        Orientation of extended trapezoidal gradient event. Must be one of x, y or z.
    times : numpy.ndarray, optional
        Time points at which `amplitudes` defines amplitude values. Default is 0.
    amplitudes : numpy.ndarray, optional
        Values defined at `times` time indices. Default is 0.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    max_grad : float, optional
        Maximum gradient strength. Default is 0.
    max_slew : float, optional
        Maximum slew rate. Default is 0.
    skip_check : bool, optional
        Perform check. Default is false.

    Returns
    -------
    grad : SimpleNamespace
        Extended trapezoid gradient event.
    """
    if channel not in ['x', 'y', 'z']:
        raise ValueError()

    if not np.any(times):
        raise ValueError('At least one of the given times must be non-zero')

    if np.any(np.diff(times) <= 0):
        raise ValueError('Times must be in ascending order and all times must be distinct')

    if not np.any(amplitudes):
        raise ValueError('At least one of the given amplitudes must be non-zero')

    if skip_check is False and times[0] > 0 and amplitudes[0] != 0:
        raise ValueError('If first amplitude of a gradient is non-zero, it must connect to previous block')

    if max_grad <= 0:
        max_grad = system.max_grad

    if max_slew <= 0:
        max_slew = system.max_slew

    waveform = points_to_waveform(times=times, amplitudes=amplitudes, grad_raster_time=system.grad_raster_time)
    grad = make_arbitrary_grad(channel=channel, waveform=waveform, system=system, max_grad=max_grad, max_slew=max_slew,
                               delay=times[0])
    grad.first = amplitudes[0]
    grad.last = amplitudes[-1]

    return grad



from types import SimpleNamespace

import numpy as np

from pypulseq.opts import Opts


def make_arbitrary_grad(channel: str, waveform: np.ndarray, system: Opts = Opts(), max_grad: float = 0,
                        max_slew: float = 0, delay: float = 0) -> SimpleNamespace:
    """
    Creates a gradient event with arbitrary waveform.

    Parameters
    ----------
    channel : str
        Orientation of gradient event of arbitrary shape. Must be one of `x`, `y` or `z`.
    waveform : numpy.ndarray
        Arbitrary waveform.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
     max_grad : float, optional
        Maximum gradient strength. Default is 0.
     max_slew : float, optional
        Maximum slew rate. Default is 0.
    delay : float, optional
        Delay in milliseconds (ms).

    Returns
    -------
    grad : SimpleNamespace
        Gradient event with arbitrary waveform.
    """
    if channel not in ['x', 'y', 'z']:
        raise ValueError(f'Invalid channel. Must be one of x, y or z. You passed: {channel}')

    if max_grad <= 0:
        max_grad = system.max_grad

    if max_slew <= 0:
        max_slew = system.max_slew

    g = waveform
    slew = np.squeeze(np.subtract(g[1:], g[:-1]) / system.grad_raster_time)
    if max(abs(slew)) >= max_slew:
        raise ValueError(f'Slew rate violation {max(abs(slew)) / max_slew * 100}')
    if max(abs(g)) >= max_grad:
        raise ValueError(f'Gradient amplitude violation {max(abs(g)) / max_grad * 100}')

    grad = SimpleNamespace()
    grad.type = 'grad'
    grad.channel = channel
    grad.waveform = g
    grad.delay = delay
    grad.t = np.arange(len(g)) * system.grad_raster_time
    grad.first = g[0]
    grad.last = g[-1]

    return grad



import math

import numpy as np

from pypulseq.opts import Opts
from pypulseq.utils import k2g


def vds_2d(fov: float, N: float, n_shots: int, alpha: int, system: Opts):
    """
    Generates a variable density k-space trajectory spiral with a method adapted from [1].

    [1] "Simple Analytic Variable Density Spiral Design", Dong-hyun Kim, Elfar Adalsteinsson, and Daniel M. Spielman,
    Magnetic Resonance in Medicine 50:214-219 (2003).

    Parameters
    ----------
    fov : float
        Field of view in meters.
    N : int
        Resolution (Nyquist distance) in meters.
    n_shots : int
        Number of interleaves.
    alpha : int
        Variable density factor.
    system : Opts
        System limits.

    Returns
    -------
    k_shot : numpy.ndarray
        K-space trajectory for the nth shot.
    Gn : numpy.ndarray
        Gradient waveform for the nth shot.
    lamda : float
        Inter-shot distance factor.
    """

    gamma = 42576000
    max_grad = system.max_grad / gamma
    slew_safety = 0.8
    max_slew = system.max_slew / gamma * slew_safety
    res = fov / N
    lamda = 0.5 / res
    n = (1 / (1 - pow(1 - n_shots / fov / lamda, 1 / alpha)))
    w = 2 * math.pi * n
    tea = lamda * w / gamma / max_grad / (alpha + 1)
    tes = math.sqrt(lamda * pow(w, 2) / max_slew / gamma) / (alpha / 2 + 1)
    ts2a = pow((pow(tes, (alpha + 1) / (alpha / 2 + 1)) * (alpha / 2 + 1) / tea / (alpha + 1)), (1 + 2 / alpha))

    if ts2a < tes:
        tau_trans = pow(ts2a / tes, 1 / (alpha / 2 + 1))
        tau = lambda t: pow(t / tes, 1 / (alpha / 2 + 1)) * (0 <= t) * (t <= ts2a) + pow(
            (((t - ts2a) / tea) + pow(tau_trans, alpha + 1)), 1 / (alpha + 1)) * (t > ts2a) * (t <= tea) * (tes >= ts2a)
        t_end = tea
    else:
        tau = lambda t: pow((t / tes), (1 / (alpha / 2 + 1)) * (0 <= t) * (t <= tes))
        t_end = tes

    k = lambda t: lamda * tau(t) ** alpha * pow(math.e, 1j * w * tau(t))
    dt = tea * 1e-4
    Dt = dt / fov / abs(k(tea) - k(tea - dt))
    t = np.arange(0, t_end, Dt)
    kt = k(t)

    DT = system.grad_raster_time
    tn = np.arange(0, t_end, DT)
    kn = k(tn)

    ktn = np.array([np.real(kn), np.imag(kn)])
    Gx, Gy, _ = k2g.k2g(ktn, DT)

    k_shot = np.zeros((len(kn), n_shots)).astype(np.complex)
    Gn = np.zeros((len(kn), n_shots)).astype(np.complex)
    complex_vectorized = np.vectorize(lambda x, y: complex(x, y))
    for s in range(n_shots):
        k_shot[:, s] = kn * pow(math.e, 2 * math.pi * 1j * (s + 1) / n_shots)
        km = np.array([np.real(np.squeeze(k_shot[:, s])), np.imag(np.squeeze(k_shot[:, s]))])
        Gx, Gy, _ = k2g.k2g(km, DT)
        Gn[:, s] = complex_vectorized(Gx, Gy)

    return k_shot, Gn, lamda



import setuptools

from pypulseq import major, minor, revision

try:  # Unicode decode error on Windows
    with open("README.md", "r") as fh:
        long_description = fh.read()
except:
    long_description = 'Pulseq in Python'

setuptools.setup(
    name="pypulseq",
    version=".".join([major, minor, revision]),
    author="Keerthi Sravan Ravi",
    author_email="ks3621@columbia.edu",
    description="Pulseq in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/imr-framework/pypulseq",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'pypulseq.utils.SAR': ['QGlobal.mat']},
    install_requires=['matplotlib>=3.0.3',
                      'numpy>=1.16.3',
                      'scipy>=1.4.1'],
    license='License :: OSI Approved :: GNU Affero General Public License v3',
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
)



# Copyright of the Board of Trustees of Columbia University in the City of New York

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import numpy.matlib
import scipy.io as sio
from scipy import interpolate

from pypulseq.Sequence.sequence import Sequence
from pypulseq.calc_duration import calc_duration


def __calc_SAR(Q, I):
    """
        Compute the SAR output for a given Q matrix and I current values.

        Parameters
        ----------
        Q : numpy.ndarray
        I : numpy.ndarray

        Returns
        -------
        SAR : numpy.ndarray
           contains the SAR value for a particular Q matrix
        """

    if len(I.shape) == 1:  # Just to fit the multi-transmit case for now, ToDo
        I = np.matlib.repmat(I, Q.shape[0], 1)  # Nc x Nt

    Ifact = np.divide(np.matmul(I, np.matrix(I).getH()), I.shape[1])
    SAR_temp = np.multiply(Q, Ifact)
    SAR = np.abs(np.sum(SAR_temp[:]))
    return SAR


def __loadQ():
    """
    Load Q matrix that is precomputed based on the VHM model for 8 channels.

    Returns
    -------
    Qtmf, Qhmf : numpy.ndarray
        Contains the Q-matrix, GSAR head and body for now.
    """

    # Load relevant Q matrices computed from the model - this code will be integrated later - starting from E fields
    Qpath = str(Path(__file__).parent / 'QGlobal.mat')
    Qmat = sio.loadmat(Qpath)
    Q = Qmat['Q']
    val = Q[0, 0]

    Qtmf = val['Qtmf']
    Qhmf = val['Qhmf']
    return Qtmf, Qhmf


def __SAR_from_seq(seq, Qtmf, Qhmf):
    """
    Compute global whole body and head only SAR values.

    Parameters
    ----------
    seq : Sequence
    Qtmf : numpy.ndarray
    Qhmf : numpy.ndarray

    Returns
    -------
    SAR_wbg_vec : numpy.ndarray
    SAR_hg_vec : numpy.ndarray
    t_vec : numpy.ndarray
        Contains the Q-matrix, GSAR head and body for now.
    """

    # Identify RF blocks and compute SAR - 10 seconds must be less than twice and 6 minutes must be less than
    # 4 (WB) and 3.2 (head-20)
    block_events = seq.block_events
    num_events = len(block_events)
    t_vec = np.zeros(num_events)
    SAR_wbg_vec = np.zeros(t_vec.shape)
    SAR_hg_vec = np.zeros(t_vec.shape)
    t_prev = 0

    for iB in block_events:
        block = seq.get_block(iB)
        block_dur = calc_duration(block)
        t_vec[iB - 1] = t_prev + block_dur
        t_prev = t_vec[iB - 1]
        if hasattr(block, 'rf'):  # has rf
            rf = block.rf
            t = rf.t
            signal = rf.signal
            # This rf could be parallel transmit as well
            SAR_wbg_vec[iB] = __calc_SAR(Qtmf, signal)
            SAR_hg_vec[iB] = __calc_SAR(Qhmf, signal)

    return SAR_wbg_vec, SAR_hg_vec, t_vec


def __SAR_interp(SAR, t):
    """
    Interpolate SAR values for one second resolution.

    Parameters
    ----------
    SAR : numpy.ndarray
    t : numpy.ndarray

    Returns
    -------
    __SAR_interp : numpy.ndarray
    tsec : numpy.ndarray
        Interpolated values of SAR for a temporal resolution of 1 second
    """
    tsec = np.arange(1, np.floor(t[-1]) + 1, 1)
    f = interpolate.interp1d(t, SAR)
    SARinterp = f(tsec)
    return SARinterp, tsec


def __SAR_lims_check(SARwbg_lim_s, SARhg_lim_s, tsec):
    """
    Check for SAR violations as compared to IEC 10 second and 6 minute averages;
    returns SAR values that are interpolated for the fixed IEC time intervals.

    Parameters
    ----------
    SARwbg_lim_s : numpy.ndarray
    SARhg_lim_s : numpy.ndarray
    tsec : numpy.ndarray

    Returns
    -------
    SAR_wbg_tensec : numpy.ndarray
    SAR_wbg_sixmin : numpy.ndarray
    SAR_hg_tensec : numpy.ndarray
    SAR_hg_sixmin : numpy.ndarray
    SAR_wbg_sixmin_peak : numpy.ndarray
    SAR_hg_sixmin_peak : numpy.ndarray
    SAR_wbg_tensec_peak : numpy.ndarray
    SAR_hg_tensec_peak : numpy.ndarray
    """
    if tsec[-1] > 10:
        SixMinThresh_wbg = 4
        TenSecThresh_wbg = 8

        SixMinThresh_hg = 3.2
        TenSecThresh_hg = 6.4

        SARwbg_lim_app = np.concatenate((np.zeros(5), SARwbg_lim_s, np.zeros(5)), axis=0)
        SARhg_lim_app = np.concatenate((np.zeros(5), SARhg_lim_s, np.zeros(5)), axis=0)

        SAR_wbg_tensec = __do_sw_sar(SARwbg_lim_app, tsec, 10)  # < 2  SARmax
        SAR_hg_tensec = __do_sw_sar(SARhg_lim_app, tsec, 10)  # < 2 SARmax
        SAR_wbg_tensec_peak = np.round(np.max(SAR_wbg_tensec), 2)
        SAR_hg_tensec_peak = np.round(np.max(SAR_hg_tensec), 2)

        if ((np.max(SAR_wbg_tensec) > TenSecThresh_wbg) or (np.max(SAR_hg_tensec) > TenSecThresh_hg)):
            print('Pulse exceeding 10 second Global SAR limits, increase TR')
        SAR_wbg_sixmin = 'NA'
        SAR_hg_sixmin = 'NA'
        SAR_wbg_sixmin_peak = 'NA'
        SAR_hg_sixmin_peak = 'NA'

        if tsec[-1] > 600:
            SARwbg_lim_app = np.concatenate((np.zeros(300), SARwbg_lim_s, np.zeros(300)), axis=0)
            SARhg_lim_app = np.concatenate((np.zeros(300), SARhg_lim_s, np.zeros(300)), axis=0)

            SAR_hg_sixmin = __do_sw_sar(SARhg_lim_app, tsec, 600)
            SAR_wbg_sixmin = __do_sw_sar(SARwbg_lim_app, tsec, 600)
            SAR_wbg_sixmin_peak = np.round(np.max(SAR_wbg_sixmin), 2)
            SAR_hg_sixmin_peak = np.round(np.max(SAR_hg_sixmin), 2)

            if ((np.max(SAR_hg_sixmin) > SixMinThresh_wbg) or (np.max(SAR_hg_sixmin) > SixMinThresh_hg)):
                print('Pulse exceeding 10 second Global SAR limits, increase TR')
    else:
        print('Need at least 10 seconds worth of sequence to calculate SAR')
        SAR_wbg_tensec = 'NA'
        SAR_wbg_sixmin = 'NA'
        SAR_hg_tensec = "NA"
        SAR_hg_sixmin = "NA"
        SAR_wbg_sixmin_peak = 'NA'
        SAR_hg_sixmin_peak = 'NA'
        SAR_wbg_tensec_peak = 'NA'
        SAR_hg_tensec_peak = 'NA'

    return SAR_wbg_tensec, SAR_wbg_sixmin, SAR_hg_tensec, SAR_hg_sixmin, SAR_wbg_sixmin_peak, SAR_hg_sixmin_peak, SAR_wbg_tensec_peak, SAR_hg_tensec_peak


def __do_sw_sar(SAR, tsec, t):
    """
    Compute a sliding window average of SAR values.

    Parameters
    ----------
    SAR : numpy.ndarray
    tsec : numpy.ndarray
    t : numpy.ndarray

    Returns
    -------
    SAR_timeavag : numpy.ndarray
        Sliding window time average of SAR values
    """
    SAR_timeavg = np.zeros(len(tsec) + int(t))
    for instant in range(int(t / 2), int(t / 2) + (int(tsec[-1]))):  # better to go from  -sw / 2: sw / 2
        SAR_timeavg[instant] = sum(SAR[range(instant - int(t / 2), instant + int(t / 2) - 1)]) / t
    SAR_timeavg = SAR_timeavg[int(t / 2):int(t / 2) + (int(tsec[-1]))]
    return SAR_timeavg


def calc_SAR(seq):
    """
    Compute Global SAR values on the `seq` object for head and whole body over the specified time averages.

    Parameters
    ----------
    seq : Sequence
        pypulseq `Sequence` object for which global SAR values will be computed.
    """

    if isinstance(seq, str):
        seq_obj = Sequence()
        seq_obj.read(seq)
        seq = seq_obj

    Qtmf, Qhmf = __loadQ()
    SARwbg, SARhg, t_vec = __SAR_from_seq(seq, Qtmf, Qhmf)
    SARwbg_lim, tsec = __SAR_interp(SARwbg, t_vec)
    SARhg_lim, tsec = __SAR_interp(SARhg, t_vec)
    SAR_wbg_tensec, SAR_wbg_sixmin, SAR_hg_tensec, SAR_hg_sixmin, SAR_wbg_sixmin_peak, SAR_hg_sixmin_peak, SAR_wbg_tensec_peak, SAR_hg_tensec_peak = __SAR_lims_check(
        SARwbg_lim, SARhg_lim, tsec)

    # Plot 10 sec average SAR
    if (tsec[-1] > 10):
        plt.plot(tsec, SAR_wbg_tensec, label='Whole Body: 10sec')
        plt.plot(tsec, SAR_hg_tensec, label='Head only: 10sec')

        # plt.plot(t_vec, SARwbg, label='Whole Body - instant')
        # plt.plot(t_vec, SARhg, label='Whole Body - instant')

        plt.xlabel('Time (s)')
        plt.ylabel('SAR (W/kg)')
        plt.title('Global SAR  - Mass Normalized -  Whole body and head only')
        ax = plt.subplot(111)

        # ax.legend()
        plt.grid(True)
        plt.show()



from types import SimpleNamespace

import numpy as np

from pypulseq.calc_duration import calc_duration
from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.opts import Opts


def split_gradient(grad: np.ndarray, system: Opts = Opts()):
    """
    Split gradient waveform `grad` into two gradient waveforms at the center.

    Parameters
    ----------
    grad : numpy.ndarray
        Gradient waveform to be split into two gradient waveforms.
    system : Opts, optional
        System limits. Default is a system limits object initialised to default values.
    Returns
    -------
    grad1, grad2 : numpy.ndarray
        Split gradient waveforms.
    """
    grad_raster_time = system.grad_raster_time
    total_length = calc_duration(grad)

    if grad.type == 'trap':
        ch = grad.channel
        grad.delay = round(grad.delay / grad_raster_time) * grad_raster_time
        grad.rise_time = round(grad.rise_time / grad_raster_time) * grad_raster_time
        grad.flat_time = round(grad.flat_time / grad_raster_time) * grad_raster_time
        grad.fall_time = round(grad.fall_time / grad_raster_time) * grad_raster_time

        times = [0, grad.rise_time]
        amplitudes = [0, grad.amplitude]
        ramp_up = make_extended_trapezoid(channel=ch, system=system, times=times, amplitudes=amplitudes,
                                          skip_check=True)
        ramp_up.delay = grad.delay

        times = [0, grad.fall_time]
        amplitudes = [grad.amplitude, 0]
        ramp_down = make_extended_trapezoid(channel=ch, system=system, times=times, amplitudes=amplitudes,
                                            skip_check=True)
        ramp_down.delay = total_length - grad.fall_time
        ramp_down.t = ramp_down.t * grad_raster_time

        flat_top = SimpleNamespace()
        flat_top.type = 'grad'
        flat_top.channel = ch
        flat_top.delay = grad.delay + grad.rise_time
        flat_top.t = np.arange(step=grad_raster_time,
                               stop=ramp_down.delay - grad_raster_time - grad.delay - grad.rise_time)
        flat_top.waveform = grad.amplitude * np.ones(len(flat_top.t))
        flat_top.first = grad.amplitude
        flat_top.last = grad.amplitude

        return ramp_up, flat_top, ramp_down
    elif grad.type == 'grad':
        raise ValueError('Splitting of arbitrary gradients is not implemented yet.')
    else:
        raise ValueError('Splitting of unsupported event.')
